#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
BETSE-specific `symlink` subcommands for `setuptools`.

## Microsoft Windows

Microsoft Windows does _not_ comply with POSIX standards and hence does _not_
generally support symbolic links.

While post-Vista versions of Microsoft Windows _do_ purport to support symbolic
links, the Windows version of the (Ana|Mini)conda Python distribution does _not_
appear to (at least, not reliably). Since this renders symbolic links useless
for standard Windows use, this module assumes Windows to _never_ support
symbolic links regardless of version.

Under Microsoft Windows, this module "fakes" symbolic link-based installation by
artifically prepending the Python-specific `sys.path` list of search dirnames
with the absolute path of the parent directory containing the top-level Python
package -- which largely has the same effect, albeit less resiliently. While
`sys.path` manipulation is (justifiably) frowned upon, no alternatives exist.
'''

# ....................{ IMPORTS                            }....................
import os
from betse_setup import buputil
from setuptools.command.install import install
from setuptools.command.install_lib import install_lib
from setuptools.command.install_scripts import install_scripts
from os import path

# ....................{ COMMANDS                           }....................
def add_setup_commands(metadata: dict, setup_options: dict) -> None:
    '''
    Add `symlink` subcommands to the passed dictionary of `setuptools` options.
    '''

    buputil.add_setup_command_classes(
        metadata, setup_options,
        symlink, symlink_lib, symlink_scripts, unsymlink)

# ....................{ CLASSES ~ install                  }....................
class symlink(install):
    '''
    Editably install (e.g., in a symbolically linked manner) `betse` into the
    active Python 3 interpreter *without* performing dependency resolution.

    Unlike the default `develop` command, this command is suitable for
    system-wide installation.
    '''

    # ..................{ ATTRIBUTES                         }..................
    description = (
        'install a symlink rather than copy of this package (for development)')
    '''
    Command description printed when running `./setup.py --help-commands`.
    '''


    sub_commands = [
        ('symlink_lib', None),
        ('symlink_scripts', None),
    ]
    '''
    Dictionary mapping command names to either:

    * A predicate returning a boolean indicating whether such command should be
      run under this run of the current command.
    * `None` signifying that such command should always be run.
    '''

    # ..................{ SUPERCLASS                         }..................
    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., `install`).
        '''

        super().finalize_options()

        #FIXME: Replicate this functionality for the "install" command as well.

        # If the current system is OS X *AND* the OS X-specific Homebrew package
        # manager is installed...
        if buputil.is_os_os_x() and buputil.is_pathable('brew'):
             # Absolute path of Homebrew's top-level system-wide cellar
             # directory (e.g., "/usr/local/Cellar").
             brew_cellar_dir = buputil.get_command_output('brew', '--cellar')
             #print('Here!')

             # Absolute path of Homebrew's top-level system-wide directory
             # (e.g., "/usr/local").
             brew_dir = buputil.get_command_output('brew', '--prefix')

             # Absolute path of Homebrew's top-level system-wide binary
             # directory (e.g., "/usr/local/bin").
             brew_binary_dir = path.join(brew_dir, 'bin')

             # If this directory does not exist, raise an exception.
             buputil.die_unless_dir(brew_binary_dir)

             # If the directory to which wrappers will be installed is a Python-
             # specific subdirectory of this cellar directory (e.g.,
             # "/usr/local/Cellar/python3/3.5.0/Frameworks/Python.framework/Versions/3.5/bin"),
             # that subdirectory is unlikely to reside in the current ${PATH},
             # in which case wrappers installed to that subdirectory will remain
             # inaccessible. Correct this by forcing wrappers to be installed
             # to the Homebrew's conventional binary directory instead.
             if self.install_scripts.startswith(brew_cellar_dir):
                 self.install_scripts = brew_binary_dir
                 print('Detected Homebrew installation directory "{}".'.format(
                     brew_binary_dir))


    def run(self):
        '''Run the current command and all subcommands thereof.'''
        # If the current operating system is POSIX-incompatible, this system
        # does *NOT* support conventional symbolic links. See details above.
        if not buputil.is_os_posix():
            # Avoid circular import dependencies.
            from betse_setup import build

            # Print a non-fatal warning.
            buputil.output_warning(
                'Symbolic links require POSIX compatibility. '
                'Since the current platform is\n'
                'POSIX-incompatible (e.g., Windows), '
                'symbolic links will be faked with black magic.'
            )

            # Absolute path of the parent directory containing the top-level
            # "betse" package.
            parent_dirname = buputil.get_project_dirname()
            # print('parent: ' + parent_dirname)

            # Prepend the template for subsequently installed entry points by a
            # Python statement "faking" symlink-based installation.
            build.SCRIPT_TEMPLATE = """
# The current operating system is POSIX-incompatible and hence does *NOT*
# support symlinks. To "fake" symlink-based installation, the standard list of
# search dirnames is prepended by the absolute path of the parent directory of
# the top-level "betse" package. For compatibility with third-party modules,
# the first entry of such list (i.e., the parent directory of this script) is
# preserved by inserting at index 1 rather than 0.
import sys
sys.path.insert(1, {})
            """.format(repr(parent_dirname)) + build.SCRIPT_TEMPLATE

        # Run all subcommands.
        for subcommand_name in self.get_sub_commands():
            self.run_command(subcommand_name)

# ....................{ CLASSES ~ install : subcommands    }....................
class symlink_lib(install_lib):
    '''
    Install the symbolic link for `betse`'s current editable installation,
    usually to a system-wide `site-packages` directory for the active Python 3
    interpreter.
    '''

    description = "install a symlink to this package's top-level module"
    '''
    Command description printed when running `./setup.py --help-commands`.
    '''


    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., `symlink`).
        '''
        # Copy attributes from a temporarily instantiated "symlink" object into
        # the current object under different attribute names.
        self.set_undefined_options(
            'symlink', ('install_lib', 'install_dir'))

        # Default all remaining options.
        super().finalize_options()


    def run(self):
        # If the current operating system is POSIX-incompatible, such system
        # does *NOT* support conventional symbolic links. Return immediately.
        if not buputil.is_os_posix():
            return

        # Absolute path of betse's top-level Python package in the current
        # directory.
        package_dirname = path.join(os.getcwd(), self._setup_options['name'])

        # Absolute path of such symbolic link.
        symlink_filename = path.join(
            self.install_dir,
            self._setup_options['name'])

        # (Re)create such link.
        buputil.make_symlink(package_dirname, symlink_filename)


class symlink_scripts(install_scripts):
    '''
    Install all scripts wrapping `betse`'s current editable installation,
    usually to a system-wide directory in the current `${PATH}`.
    '''

    description =\
        'install scripts running this package without dependency checks'
    '''
    Command description printed when running `./setup.py --help-commands`.
    '''


    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., `symlink`).
        '''
        # Copy attributes from a temporarily instantiated "symlink" object into
        # the current object under different attribute names.
        self.set_undefined_options(
            'build',
            ('build_scripts', 'build_dir'),
        )
        self.set_undefined_options(
            'symlink',
            ('install_scripts', 'install_dir'),
            ('force', 'force'),
            ('skip_build', 'skip_build'),
        )

        # Default all remaining options.
        super().finalize_options()

# ....................{ UNINSTALLERS                       }....................
class unsymlink(install):
    '''
    Editably uninstall (e.g., in a symbolically linked manner) `betse` from
    the active Python 3 interpreter.

    Attributes
    ----------
    install_package_dirname : str
        Absolute path of the directory to which our Python codebase was
        previously installed.
    install_wrapper_dirname : str
        Absolute path of the directory to which our wrapper scripts were
        previously installed.
    '''

    description =\
        'uninstall all installed symbolic links and scripts for this package'
    '''
    Command description printed when running `./setup.py --help-commands`.
    '''


    def initialize_options(self):
        '''
        Declare option-specific attributes subsequently initialized by
        `finalize_options()`.

        If this function is *not* defined, the default implementation of this
        method raises an inscrutable `distutils` exception. If such attributes
        are *not* declared, the subsequent call to
        `self.set_undefined_options()` raises an inscrutable `setuptools`
        exception. (This is terrible. So much hate.)
        '''
        super().initialize_options()
        self.install_package_dirname = None
        self.install_wrapper_dirname = None


    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., `symlink`).
        '''
        # Copy attributes from a temporarily instantiated "symlink" object into
        # the current object under different attribute names.
        self.set_undefined_options(
            'symlink',
            ('install_lib',     'install_package_dirname'),
            ('install_scripts', 'install_wrapper_dirname'),
        )

        # Default all remaining options.
        super().finalize_options()


    def run(self):
        '''Run the current command and all subcommands thereof.'''
        # If the current operating system is POSIX-compatible, such system
        # supports symbolic links. In such case, remove the previously installed
        # symbolic link.
        if buputil.is_os_posix():
            buputil.remove_symlink(path.join(
                self.install_package_dirname,
                self._setup_options['name'],
            ))

        # Remove all installed scripts.
        for script_basename, _, _ in buputil.command_entry_points(self):
            buputil.remove_file(path.join(
                self.install_wrapper_dirname, script_basename))



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Create and electrodiffuses a suite of customizable general molecule in the BETSE
ecosystem, including functionality to pump the molecule, use it as a gating
ligand, produce and consume it, and use it an enzyme to facilitate another
reaction.

The molecule is assumed to be at low concentrations and to not have a
significant effect on system voltages or currents. This module creates a
structure containing all user-defined molecules, along with the facilities to
initialize, define the core computations for a simulation loop, save and report
on data, and plot.
'''

#FIXME: This submodule no longer appears to be used anywhere in the codebase.
#Can we safely remove this now? If not, no worries. Lumbering lumberjacks!

import numpy as np
from betse.science import filehandling as fh
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.science.chemistry.networks import MasterOfNetworks
from betse.science.config import confio
from betse.exceptions import BetseSimConfException
from betse.science.chemistry.netplot import set_net_opts


class MasterOfMetabolism(object):

    def __init__(self, p):

        #FIXME: Extract the "MetabolicNetwork.betse" basename into a new
        #configuration option, perhaps in the "init file saving" section.
        #FIXME: Replace "MetabolicNetwork.betse" with
        #"MetabolicNetwork.betse.gz" to compress this pickled file.

        # Define data paths for saving an initialization and simulation run:
        self.savedMoM = pathnames.join(p.init_pickle_dirname, 'MetabolicNetwork.betse')

    def read_metabo_config(self, sim, cells, p):

        # create the path to read the metabolism config file:
        self.configPath = pathnames.join(
            p.conf_dirname, p.metabo_config_filename)

        # read the config file into a dictionary:
        self.config_dic = confio.read_metabo(self.configPath)

        # determine if mitochondria are enabled:
        self.mit_enabled = self.config_dic['enable mitochondria']

        # obtain specific sub-dictionaries from the config file:
        substances_config = self.config_dic['biomolecules']
        reactions_config = self.config_dic.get('reactions', None)
        transporters_config = self.config_dic.get('transporters', None)
        channels_config = self.config_dic.get('channels', None)
        modulators_config = self.config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core = MasterOfNetworks(sim, cells, substances_config, p, mit_enabled=self.mit_enabled)

        if reactions_config is not None:

            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            if self.mit_enabled is True:
                self.core.write_reactions_mit()
                self.core.create_reaction_matrix_mit()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None and len(transporters_config) >0:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)
            self.transporters = True

        else:
            self.transporters = False

        # initialize any custom channels:-------------

        if channels_config is not None and len(channels_config) > 0:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize any modulators------------------

        if modulators_config is not None and len(modulators_config) > 0:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

        # test to make sure the metabolic simulation includes core components:
        if 'ATP' not in self.core.molecules or 'ADP' not in self.core.molecules or 'Pi' not in self.core.molecules:

            raise BetseSimConfException("This metabolic simulation does not contain key substances."
                                           "Please define 'ATP', 'ADP' and 'Pi' biomolecules in your "
                                           "metabolism configuration file and try again.")

        # read in network plotting options:
        self.core.net_plot_opts = self.config_dic.get('network plotting', None)

        # set plotting options for the network:
        set_net_opts(self.core, self.core.net_plot_opts, p)

            # after primary initialization, check and see if optimization required:
        opti = self.config_dic['optimization']['optimize network']
        self.core.opti_N = self.config_dic['optimization']['optimization steps']
        self.core.opti_method = self.config_dic['optimization']['optimization method']
        self.core.target_vmem = float(self.config_dic['optimization']['target Vmem'])
        self.core.opti_T = float(self.config_dic['optimization']['optimization T'])
        self.core.opti_step = float(self.config_dic['optimization']['optimization step'])
    #     self.core.opti_run = self.config_dic['optimization']['run from optimization']
    #
        if opti is True:
            logs.log_info("The Metabolic Network is being analyzed for optimal rates...")
            self.core.optimizer(sim, cells, p)
            self.reinitialize(sim, cells, p)
    #
    #     if self.core.opti_run is True:
    #         self.run_from_init(self, sim, cells, p)
    #
    # def run_from_init(self, sim, cells, p):
    #
    #     pass

    #FIXME: Oh, boy. Most of this method appears to have been copy-and-pasted
    #from the read_metabo_config() method above. That's... not the best. Let's
    #extract the code shared in common between these two methods into a new
    #_init_metabo() method internally called by these two methods.
    def reinitialize(self, sim, cells, p):

        # create the path to read the metabolism config file:
        self.configPath = pathnames.join(
            p.conf_dirname, p.metabo_config_filename)

        # read the config file into a dictionary:
        self.config_dic = confio.read_metabo(self.configPath)

        # determine if mitochondria are enabled:
        self.mit_enabled = self.config_dic['enable mitochondria']

        # obtain specific sub-dictionaries from the config file:
        substances_config = self.config_dic['biomolecules']
        reactions_config = self.config_dic.get('reactions', None)
        transporters_config = self.config_dic.get('transporters', None)
        channels_config = self.config_dic.get('channels', None)
        modulators_config = self.config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core.tissue_init(sim, cells, substances_config, p)

        if reactions_config is not None and len(reactions_config):

            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            if self.mit_enabled is True:
                self.core.write_reactions_mit()
                self.core.create_reaction_matrix_mit()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None and len(transporters_config) > 0:
            self.core.read_transporters(transporters_config, sim, cells, p)
            self.core.write_transporters(sim, cells, p)
            self.transporters = True

        else:
            self.transporters = False

        # initialize any custom channels:-------------

        if channels_config is not None and len(channels_config) > 0:
            self.core.read_channels(channels_config, sim, cells, p)
            self.channels = True

        else:
            self.channels = False

        # initialize any modulators------------------

        if modulators_config is not None and len(modulators_config) > 0:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

        # test to make sure the metabolic simulation includes core components:
        if 'ATP' not in self.core.molecules or 'ADP' not in self.core.molecules or 'Pi' not in self.core.molecules:
            raise BetseSimConfException("This metabolic simulation does not contain key substances."
                                           "Please define 'ATP', 'ADP' and 'Pi' biomolecules in your "
                                           "metabolism configuration file and try again.")

    def run_core_sim(self, sim, cells, p):
        """
        Runs a simulation of the biochemical reaction network only, with a dummy sim and dyna module.
        This allows the user to test the reaction network without the influence of bioelectrical dynamics.

        This method is called in the BETSE CLI command betse sim-brn my_yaml.yaml and data plotted via
        betse plot sim-brn my_yaml.yaml

        """

        # # point sim.metabo to this object
        # sim.metabo = self

        # create a dictionary pointing to key metabolic molecules used in sim: ATP, ADP and Pi:
        sim.met_concs = {'cATP': self.core.cell_concs['ATP'][cells.mem_to_cells],
            'cADP': self.core.cell_concs['ADP'][cells.mem_to_cells],
            'cPi': self.core.cell_concs['Pi'][cells.mem_to_cells]}

        # initialize Vmem to an initial value common to many cell types:
        sim.vm = -50e-3*np.ones(sim.mdl)

        sim.conc_J_x = np.zeros(sim.edl)
        sim.conc_J_y = np.zeros(sim.edl)

        # initialize key fields of simulator required to interface (dummy init)
        sim.rho_pump = 1.0
        sim.rho_channel = 1.0

        sim.J_env_x = np.zeros(sim.edl)
        sim.J_env_y = np.zeros(sim.edl)

        # p.substances_affect_charge = False

        # specify a time vector
        loop_time_step_max = p.init_tsteps
        # Maximum number of seconds simulated by the current run.
        loop_seconds_max = loop_time_step_max * p.dt
        # Time-steps vector appropriate for the current run.
        tt = np.linspace(0, loop_seconds_max, loop_time_step_max)

        tsamples = set()
        i = 0
        while i < len(tt) - p.t_resample:
            i = int(i + p.t_resample)
            tsamples.add(tt[i])

        self.core.clear_cache()
        self.time = []
        sim.time = []

        for t in tt:

            if self.transporters:
                self.core.run_loop_transporters(t, sim, cells, p)

            # if self.channels:
            #     self.core.run_loop_channels(sim, self.core, cells, p)
            #
            # if self.modulators:
            #     self.core.run_loop_modulators(sim, self.core, cells, p)

            self.core.run_loop(t, sim, cells, p)

            # # update core ions in sim:
            # for i in sim.movingIons:
            #
            #     # update the ion concentration intra-cellularly:
            #     sim.cc_mems[i][:], sim.cc_cells[i][:], _ = \
            #         stb.update_intra(sim, cells, sim.cc_mems[i][:],
            #             sim.cc_cells[i][:],
            #             sim.D_free[i],
            #             sim.zs[i], p)


            if t in tsamples:

                sim.time.append(t)

                logs.log_info('------------------' + str(np.round(t,3)) +' s --------------------')
                self.time.append(t)
                self.core.write_data(sim, cells, p)
                self.core.report(sim, p)

        logs.log_info('Saving simulation...')
        datadump = [self, cells, p]
        fh.saveSim(self.savedMoM, datadump)
        self.core.init_saving(cells, p, plot_type='init', nested_folder_name='Metabolism')
        self.core.export_eval_strings(p)
        self.core.export_equations(p)
        message = 'Metabolic network simulation saved to' + ' ' + self.savedMoM
        logs.log_info(message)

        logs.log_info('-------------------Simulation Complete!-----------------------')

    def update_ATP(self, flux, sim, cells, p):

        """
        Update ATP, ADP and Pi concentrations using a
        concentration change defined on membranes or cell
        centres.

        This method is typically called in sim after ATPase pumps are run.

        flux        concentration flux [mol/m2 s], negative consumes ATP; array must be of length sim.mdl

        """

        cATP = self.core.cell_concs['ATP']
        cADP = self.core.cell_concs['ADP']
        cPi = self.core.cell_concs['Pi']

        deltac = (np.dot(cells.M_sum_mems, flux*cells.mem_sa)/cells.cell_vol)*p.dt

        self.core.cell_concs['ATP'] = cATP + deltac
        self.core.cell_concs['ADP'] = cADP - deltac
        self.core.cell_concs['Pi'] = cPi - deltac

        sim.met_concs = {'cATP': self.core.cell_concs['ATP'][cells.mem_to_cells],
            'cADP': self.core.cell_concs['ADP'][cells.mem_to_cells],
            'cPi': self.core.cell_concs['Pi'][cells.mem_to_cells]}



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Metadata constants synopsizing high-level application dependencies.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.util.type import modules

# ....................{ LIBS ~ runtime : mandatory         }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Changes to this subsection *MUST* be synchronized with:
# * Front-facing documentation (e.g., "doc/md/INSTALL.md").
# * The "betse.util.type.modules.DISTUTILS_PROJECT_NAME_TO_MODULE_NAME"
#   dictionary, converting between the setuptools-specific names listed below
#   and the Python-specific module names imported by this application.
# * Gitlab-CI configuration (e.g., the top-level "requirements-conda.txt" file).
# * Third-party platform-specific packages (e.g., Gentoo Linux ebuilds).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

RUNTIME_MANDATORY = {
    # setuptools is currently required at both install and runtime. At runtime,
    # setuptools is used to validate that dependencies are available.
    'setuptools': '>= 3.3',

    # Dependencies directly required by this application. Notably:
    #
    # * matplotlib >= 2.1.0 requires Numpy >= 1.7.1.
    # * SciPy >= 1.0.0 requires Numpy >= 1.8.2.
    #
    # SciPy wins.
    'Numpy':  '>= 1.8.2',
    'Pillow': '>= 2.3.0',
    'SciPy':  '>= 0.12.0',
    'dill':   '>= 0.2.3',

    # Matplotlib >= 1.5.0 is required for the newly added "viridis" colormap.
    'matplotlib': '>= 1.5.0',

    # Dependencies indirectly required by this application but only optionally
    # required by dependencies directly required by this application. Since the
    # "setup.py" scripts for the latter do *NOT* list these dependencies as
    # mandatory, these dependencies *MUST* be explicitly listed here.

    # Dependencies directly required by dependencies directly required by this
    # application. While these dependencies need *NOT* be explicitly listed
    # here, doing so improves detection of missing dependencies in a
    # human-readable manner.
    'six': '>= 1.5.2',       # required by everything that should not be
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
mandatory runtime dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

To simplify subsequent lookup, these dependencies are contained by a dictionary
rather than a simple set or sequence such that each:

* Key is the name of a :mod:`setuptools`-specific project identifying this
  dependency, which may have no relation to the name of that project's top-level
  module or package (e.g., the ``PyYAML`` project's top-level package is
  :mod:`yaml`). For human readability in error messages, this name should
  typically be case-sensitively capitalized -- despite being parsed
  case-insensitively by :mod:`setuptools`.
* Value is either:
  * ``None`` or the empty string, in which case this dependency is unconstrained
    (i.e., any version of this dependency is sufficient).
  * A string of the form ``{comparator} {version}``, where:
    * ``{comparator}`` is a comparison operator (e.g., ``>=``, ``!=``).
    * ``{version}`` is the required version of this project to compare.

Concatenating each such key and value yields a :mod:`setuptools`-specific
requirements string of the form either ``{project_name}`` or ``{project_name}
{comparator} {version}``.

Official :mod:`setuptools` documentation suggests the ``install_requires`` and
``setup_requires`` keys of the ``setup()`` packaging function to accept only
sequences rather than dictionaries of strings. While undocumented, these keys
*do* actually accept both sequences and dictionaries of strings.

Caveats
----------
This application requires :mod:`setuptools` at both installation time *and*
runtime -- in the latter case, to validate all application dependencies at
runtime. Note that doing so technically only requires the :mod:`pkg_resources`
package installed with :mod:`setuptools` rather than the :mod:`setuptools`
package itself. Since there exists no means of asserting a dependency on only
:mod:`pkg_resources`, however, :mod:`setuptools` is depended upon instead.

See Also
----------
https://setuptools.readthedocs.io/en/latest/setuptools.html#id12
    Further details on the :mod:`setuptools` string format for dependencies.
:download:`/doc/md/INSTALL.md`
    Human-readable list of these dependencies.
'''

# ....................{ LIBS ~ runtime : mandatory : yaml  }....................
RUNTIME_MANDATORY_YAML_PROJECT_NAME = None
'''
:mod:`setuptools`-specific project name of the current YAML implementation
satisfying this application's mandatory runtime YAML dependency.

This project is dynamically selected at submodule importation time based on the
most preferred implementation currently importable from the
:data:`RUNTIME_MANDATORY_YAML` dictionary.
'''


RUNTIME_MANDATORY_YAML = {
    'PyYAML': '>= 3.10',

    #FIXME: As "ruamel.yaml" now purports to safely roundtrip our documents,
    #let's give it a whirl and see how far we get plunge into the icy waters of
    #markup purgatory this time.

    # A relatively modern version of "ruamel.yaml" variants is required.
    # Specifically, this application requires:
    #
    # * At least version 0.15.24 or newer of "ruamel.yaml", which resolves a
    #   long-standing parser issue preventing overly complex YAML markup (such
    #   as ours) from being safely roundtripped:
    #   0.15.24 (2017-08-09):
    #   * (finally) fixed longstanding issue 23 (reported by Antony Sottile),
    #     now handling comment between block mapping key and value correctly
    # * The new "ruamel.yaml" API first introduced in 0.15.0. While older
    #   versions strictly adhere to the functional PyYAML-compatible API, newer
    #   versions break backward compatibility by entirely supplanting that API
    #   with a modern object-oriented approach. Supporting both isn't worth the
    #   substantial increase in maintenance debt.
    'ruamel.yaml': '>= 0.15.24',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
YAML implementation satisfying the same mandatory runtime dependency for this
application to the suffix of a :mod:`setuptools`-specific requirements string
constraining this dependency.

Since neither :mod:`distutils` nor :mod:`setuptools` support mutual exclusivity,
these alternatives are isolated into a separate dictionary implicitly merged
into the :data:`RUNTIME_MANDATORY` dictionary at submodule importation time.

Motivation
----------
Ideally, exactly one YAML implementation would be required. While
:mod:`ruamel.yaml` is the most preferred YAML implementation and hence the
obvious candidate, installation of that namespace package is sufficiently
non-trivial and frankly fragile to warrant falling back to less preferable but
more widely available alternatives.
'''

# ....................{ LIBS ~ runtime : optional          }....................
#FIXME: Should these be dependencies also be added to our "setup.py" metadata,
#perhaps as so-called "extras"? Contemplate. Consider. Devise.
RUNTIME_OPTIONAL = {
    # To simplify subsequent lookup at runtime, project names for optional
    # dependencies should be *STRICTLY LOWERCASE*. Since setuptools parses
    # project names case-insensitively, case is only of internal relevance.

    # Dependencies directly required by this application.
    'pympler': '>= 0.4.1',
    'ptpython': '>= 0.29',

    # A relatively modern version of NetworkX *EXCLUDING* 1.11, which
    # critically broke backwards compatibility by coercing use of the unofficial
    # inactive "pydotplus" PyDot fork rather than the official active "pydot"
    # PyDot project, is directly required by this application. NetworkX >= 1.12
    # reverted to supporting "pydot", thus warranting blacklisting of only
    # NetworkX 1.11. It is confusing, maybe?
    'networkx': '>= 1.8, != 1.11',
    'pydot': '>= 1.0.28',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
optional runtime dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

See Also
----------
:data:`RUNTIME_MANDATORY`
    Further details on dictionary structure.
:download:`/doc/md/INSTALL.md`
    Human-readable list of these dependencies.
:func:`get_dependencies_runtime_optional_tuple`
    Function converting this dictionary of key-value string pairs into a tuple
    of strings (e.g., within :download:`/setup.py`).
'''

# ....................{ LIBS ~ testing : mandatory         }....................
TESTING_MANDATORY = {
    # For simplicity, py.test should remain the only hard dependency for testing
    # on local machines. While our setuptools-driven testing regime optionally
    # leverages third-party py.test plugins (e.g., "pytest-xdist"), these
    # plugins are *NOT* required for simple testing.
    'pytest': '>= 2.5.0',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
mandatory testing dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

See Also
----------
:data:`RUNTIME_MANDATORY`
    Further details on dictionary structure.
:download:`/doc/md/INSTALL.md`
    Human-readable list of these dependencies.
'''

# ....................{ LIBS ~ command                     }....................
class DependencyCommand(object):
    '''
    Lightweight metadata describing a single external command required by some
    application dependency (of any type, including optional, mandatory, runtime,
    testing, or otherwise).

    Attributes
    ----------
    name : str
        Human-readable name associated with this command (e.g., ``Graphviz``).
    basename : str
        Basename of this command to be searched for in the current ``${PATH}``.
    '''

    def __init__(self, name: str, basename: str) -> None:
        self.name = name
        self.basename = basename


#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Changes to this dictionary *MUST* be synchronized with:
# * Front-facing documentation (e.g., "doc/md/INSTALL.md").
# * Gitlab-CI configuration (e.g., the top-level "requirements-conda.txt" file).
# * Third-party platform-specific packages (e.g., Gentoo Linux ebuilds).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
EXTERNAL_COMMANDS = {
    'pydot': (DependencyCommand(name='Graphviz', basename='dot'),),
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
application dependency (of any type, including optional, mandatory, runtime,
testing, or otherwise) requiring one or more external commands to a tuple of
:class:`DependencyCommand` instances describing these requirements.

See Also
----------
:download:`/doc/md/INSTALL.md`
    Human-readable list of these dependencies.
'''

# ....................{ INITIALIZERS                       }....................
def _init() -> None:
    '''
    Dynamically finalize the contents of the :data:`RUNTIME_MANDATORY` global
    declared by this submodule.

    Specifically, this function selects the first importable third-party YAML
    implementation (in descending order of preference) and merges the
    :mod:`setuptools`-specific metadata constraining this implementation from
    the :data:`RUNTIME_MANDATORY_YAML` global into the :data:`RUNTIME_MANDATORY`
    and :data:`RUNTIME_OPTIONAL` globals.

    Caveats
    ----------
    Note that this install- and runtime logic only applies to logic paths that
    actually execute this script. This excludes wheel installation, which avoids
    script execution for safety. (Because this is Python.)
    '''

    # Global variables assigned to below.
    global RUNTIME_MANDATORY_YAML_PROJECT_NAME

    # Setuptools-specific project name of the current YAML implementation
    # satisfying this application's mandatory runtime YAML dependency,
    # defaulting to the most preferred YAML framework. This has the beneficial
    # side effect of installing this framework at installation time if no other
    # YAML framework has been installed.
    RUNTIME_MANDATORY_YAML_PROJECT_NAME = 'ruamel.yaml'

    #FIXME: Handle version constraints as well. Since "setuptools" is itself a
    #non-standard dependency, handling such constraints in a portable manner
    #will probably mean leveraging "distutils" functionality for doing so.
    #Specifically, if a sufficiently new version of neither "ruamel.yaml" nor
    #"ruamel_yaml" is available, "yaml" should be fallen back to for safety.

    # Prefer "ruamel.yaml", the most actively maintained and hence preferred
    # YAML framework.
    if modules.is_module('ruamel.yaml'):
        pass
    # Fallback to "ruamel_yaml", the next most actively maintained and hence
    # preferred YAML framework. Unlike the more general-purpose "ruamel.yaml"
    # namespace package, "ruamel_yaml" is an Anaconda-specific non-namespace
    # package internally leveraged by the "conda" package manager but only
    # infrequently rebased against upstream changes.
    #FIXME: Uncomment once "ruamel_yaml" is sufficiently up-to-date.
    # elif modules.is_module('ruamel_yaml'):
    #     RUNTIME_MANDATORY_YAML_PROJECT_NAME = 'ruamel_yaml'
    # Fallback to PyYaml, the officially dead and hence least preferred YAML
    # framework.
    elif modules.is_module('yaml'):
        RUNTIME_MANDATORY_YAML_PROJECT_NAME = 'PyYAML'

    #FIXME: *DELETE THE FOLLOWING LINE AFTER* upstream "ruamel.yaml" roundtrip
    #issues are resolved. Currently, "ruamel.yaml" fails to properly roundtrip
    #our default simulation configuration and hence must *NOT* be used. *sigh*
    #
    #After upstream both resolves these issues *AND* releases a new stable
    #release, modify the "ruamel.yaml" version constraints above to reflect the
    #new minimum required version. Naturally, this implies we won't be enabling
    #"ruamel.yaml" support anytime soon.
    #FIXME: *UGH.* "ruamel.yaml" appears to have a core deficiency with respect
    #to map key comments: specifically, the roundtripper silently produces
    #malformed YAML when encountering map key comments. This is quite obviously
    #unacceptable for our heavily commented YAML files. Sadly, upstream has
    #known about this open issue for several years -- implying a fix is distant
    #at beast. Until resolved, "ruamel.yaml" *MUST* be ignored. For discussion,
    #see the following upstream issue:
    #    https://bitbucket.org/ruamel/yaml/issues/146/torture-test-breaks-the-fragile-back-of
    RUNTIME_MANDATORY_YAML_PROJECT_NAME = 'PyYAML'

    # Enforce installation of the preferred YAML framework detected above.
    RUNTIME_MANDATORY[RUNTIME_MANDATORY_YAML_PROJECT_NAME] = (
        RUNTIME_MANDATORY_YAML[RUNTIME_MANDATORY_YAML_PROJECT_NAME])

    # Merge the requirement strings for all YAML frameworks into those for all
    # optional dependencies, permitting the former to be treated like the latter
    # (e.g., by the betse.lib.libs.die_unless_runtime_optional() function).
    RUNTIME_OPTIONAL.update(RUNTIME_MANDATORY_YAML)

# ....................{ MAIN                               }....................
# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ CONVENIENCES                       }....................
#FIXME: Revise docstring. This function no longer contains test-specific logic.
def _ignite() -> None:
    '''
    Initialize both the current application _and_ all mandatory third-party
    dependencies of this application with sane defaults if this application
    both has not already been initialized _and_ is not currently being tested.

    This function is a convenience intended to streamline interactive use (e.g.,
    from web-based IPython notebooks or CLI-based Python consoles) by implicitly
    initializing this application on the first import of this subpackage.

    If this application either has already been manually initialized _or_ is
    currently being tested by a test suite that that may subsequently do so,this
    call silently reduces to a noop -- which is a good thing.
    '''

    # Localize *ALL* imports to this function, preventing the package namespace
    # from being polluted with imports requiring deletion via the del() builtin.
    from betse import ignition
    # from betse.util.py import pys

    # If tests are being run, avoid pre-initializing this application.
    # if pys.is_testing():
    #     return

    # Initialize both this application and all dependencies thereof.
    ignition.ignite()

#FIXME: Overly cumbersome. Just inline the two statements above here; then,
#excise this function.
# Initialize both this application and all dependencies thereof.
_ignite()

# ....................{ CLEANUP                            }....................
# Delete *ALL* attributes (including callables) defined above, preventing the
# package namespace from being polluted with these attributes.
del _ignite



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level classes aggregating all parameters pertaining to simulation events.
'''

# ....................{ IMPORTS                            }....................
from betse.science.tissue.event.tiseveabc import SimEventSpikeABC
# from betse.util.io.log import logs
from betse.util.type.types import type_check, SequenceTypes

# ....................{ SUBCLASSES                         }....................
#FIXME: Refactor the TissueHandler.removeCells() method into a new
#fire() method of this subclass.
class SimEventCut(SimEventSpikeABC):
    '''
    **Cutting event** (i.e., event removing a region of the current cluster at
    some time step during the simulation phase).

    Attributes
    ----------
    profile_names : list
        List of the names of all applicable cut profiles, each describing a
        subset of the cell population to be removed.
    '''

    # ..................{ PUBLIC                             }..................
    @type_check
    def __init__(
        self,
        profile_names: SequenceTypes,
        *args, **kwargs
    ) -> None:

        # Initialize our superclass with all remaining parameters..
        super().__init__(*args, **kwargs)

        # Classify all passed parameters.
        self.profile_names = profile_names



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfiguration classes for tissue and cut profiles.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.lib.yaml.yamlalias import (
    yaml_alias,
    yaml_alias_float_nonnegative,
    yaml_alias_float_percent,
    yaml_enum_alias,
)
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.lib.yaml.abc.yamllistabc import YamlList, YamlListItemABC
from betse.science.config.confenum import CellsPickerType
# from betse.util.io.log import logs
from betse.util.type.types import type_check, SequenceTypes

# ....................{ SUPERCLASSES ~ tissue              }....................
class SimConfTissueABC(object, metaclass=ABCMeta):
    '''
    Abstract mixin generalizing implementation common to all YAML-backed tissue
    profile subconfiguration subclasses.

    This mixin encapsulates configuration of a single tissue profile parsed from
    the current YAML-formatted simulation configuration file. For generality,
    this mixin provides no support for the YAML ``name`` key or the suite of
    YAML keys pertaining to tissue pickers.

    Attributes
    ----------
    name : str
        Arbitrary string uniquely identifying this tissue profile in the list of
        all tissue profiles for this simulation.

    Attributes (Membrane Diffusion)
    ----------
    Dm_Na : float
        Sodium (Na+) membrane diffusion constant in m2/s.
    Dm_K : float
        Potassium (K+) membrane diffusion constant in m2/s.
    Dm_Cl : float
        Chloride (Cl-) membrane diffusion constant in m2/s.
    Dm_Ca : float
        Calcium (Ca2+) membrane diffusion constant in m2/s.
    Dm_M : float
        Charge balance anion (M-) membrane diffusion constant in m2/s.
    Dm_P : float
        Protein (P-) membrane diffusion constant in m2/s.
    '''

    # ..................{ ALIASES                            }..................
    name = yaml_alias("['name']", str)

    # ..................{ ALIASES ~ diffusion                }..................
    Dm_Na = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Na']")
    Dm_K  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_K']")
    Dm_Cl = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Cl']")
    Dm_Ca = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Ca']")
    # Dm_H  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_H']")
    Dm_M  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_M']")
    Dm_P  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_P']")

# ....................{ SUBCLASSES ~ tissue                }....................
class SimConfTissueDefault(SimConfTissueABC, YamlABC):
    '''
    YAML-backed default tissue profile subconfiguration, encapsulating the
    configuration of a single tissue profile unconditionally applicable to all
    cells parsed from a dictionary configuring at least this profile in the
    current YAML-formatted simulation configuration file.

    Attributes (Cell Picker)
    ----------
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        define the shape of the cell cluster to be populated with cells. See
        the :attr:`SimConfTissueListItem.picker_image_filename` variable for
        details.
    '''

    # ..................{ ALIASES ~ picker                   }..................
    picker_image_filename = yaml_alias("['image']", str)


class SimConfTissueListItem(SimConfTissueABC, YamlListItemABC):
    '''
    YAML-backed tissue profile list item subconfiguration, encapsulating the
    configuration of a single tissue profile parsed from a list of these
    profiles in the current YAML-formatted simulation configuration file.

    Attributes
    ----------
    is_gj_insular : bool
        ``True`` only if gap junctions originating at cells in this tissue are
        **insular** (i.e., prevented from connecting to cells in other tissues),
        implying these gap junctions to be strictly intra-tissue.

    Attributes (Cell Picker)
    ----------
    picker_type : CellsPickerType
        Type of **tissue profile picker** (i.e., object assigning a subset of
        all cells matching some criteria to this tissue profile).
    picker_cells_index : SequenceTypes
        Sequence of the indices of all cells to be assigned to this tissue.
        Ignored unless :attr:`picker_type` is :attr:`CellsPickerType.INDICES`.
    picker_cells_percent : float
        **Percentage** (i.e., floating point number in the range ``[0.0,
        100.0]``) of the total cell population to be randomly assigned to this
        tissue. Ignored unless :attr:`picker_type` is
        :attr:`CellsPickerType.PERCENT`.
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        (i.e., pixels whose red, green, and blue color components are all 0)
        define the region of the cell cluster whose cells are all to be assigned
        to this tissue. This image *must*:
        * Be square (i.e., have equal width and height).
        * Contain no alpha transparency layer.
        Ignored unless :attr:`picker_type` is :attr:`CellsPickerType.IMAGE`.
    '''

    # ..................{ ALIASES                            }..................
    is_gj_insular = yaml_alias("['insular']", bool)

    # ..................{ ALIASES ~ picker                   }..................
    picker_type = yaml_enum_alias("['cell targets']['type']", CellsPickerType)
    picker_cells_index = yaml_alias(
        "['cell targets']['indices']", SequenceTypes)
    picker_cells_percent = yaml_alias_float_percent(
        "['cell targets']['percent']")
    picker_image_filename = yaml_alias(
        "['cell targets']['image']", str)

    # ..................{ CLASS                              }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Name of this tissue profile unique to this list.
        tissue_name = yaml_list.get_item_name_unique(
            name_format='tissue ({{}})')

        # Create and return the equivalent YAML-backed tissue profile list item,
        # duplicating the first such item in our default YAML file.
        return SimConfTissueListItem().load(conf={
            'name': tissue_name,
            'insular': True,
            'diffusion constants': {
                'Dm_Na': 1.0e-18,
                'Dm_K': 15.0e-18,
                'Dm_Cl': 2.0e-18,
                'Dm_Ca': 1.0e-18,
                # 'Dm_H': 1.0e-18,
                'Dm_M': 1.0e-18,
                'Dm_P': 0.0,
            },
            'cell targets': {
                'type': 'all',
                'image': 'geo/circle/circle_base.png',
                'indices': [3, 14, 15, 9, 265],
                'percent': 50,
            },
        })

# ....................{ SUBCLASSES ~ cut                   }....................
class SimConfCutListItem(YamlListItemABC):
    '''
    YAML-backed cut profile list item subconfiguration, encapsulating the
    configuration of a single cut profile parsed from a list of these profiles
    in the current YAML-formatted simulation configuration file.

    Attributes
    ----------
    name : str
        Arbitrary string uniquely identifying this cut profile in this list.
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        define the region of the cell cluster whose cells are all to be removed
        by this cut profile. See the
        :attr:`SimConfTissueListItem.picker_image_filename` variable for
        details.
    '''

    # ..................{ ALIASES                            }..................
    name = yaml_alias("['name']", str)
    picker_image_filename = yaml_alias("['image']", str)

    # ..................{ CLASS                              }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Name of this cut profile unique to this list.
        cut_name = yaml_list.get_item_name_unique(name_format='cut ({{}})')

        # Create and return the equivalent YAML-backed cut profile list item,
        # duplicating the first such item in our default YAML file.
        return SimConfTissueListItem().load(conf={
            'name': cut_name,
            'image': 'geo/circle/wedge.png',
        })



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level class facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseTypeException
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassType,
    GeneratorType,
    MappingType,
    SequenceTypes,
)

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_subclass(subclass: ClassType, superclass: ClassType) -> None:
    '''
    Raise an exception unless the first passed class inherits and is thus a
    subclass of the second passed class.

    Parameters
    ----------
    subclass : ClassType
        Subclass to be validated.
    superclass : ClassType
        Superclass to be validated.
    '''

    if not issubclass(subclass, superclass):
        raise BetseTypeException(
            'Class {!r} not a subclass of class {!r}'.format(
                subclass, superclass))

# ....................{ GETTERS                            }....................
@type_check
def get_module_name(cls: ClassType) -> str:
    '''
    Fully-qualified name of the module defining the passed class.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this module name for.

    Returns
    ----------
    str
        Fully-qualified name of this module.

    Raises
    ----------
    BetseTypeException
        If this class has no ``__module__`` attribute, which should ideally
        *never* happen.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # Defer to this existing function, which suffices.
    return objects.get_class_module_name(cls)

# ....................{ GETTERS ~ name                     }....................
@type_check
def get_name(cls: ClassType) -> str:
    '''
    Unqualified name of the passed class.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this unqualified name for.

    Returns
    ----------
    str
        Unqualified name of this class.
    '''

    # Elegant simplicity diminishes aggressive tendencies.
    return cls.__name__


@type_check
def get_name_snakecase(cls: ClassType) -> str:
    '''
    Unqualified name of the passed class converted from CamelCase to snake_case
    (e.g., from ``MyClassName`` to ``my_class_name``).

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this unqualified snake_case name for.

    Returns
    ----------
    str
        Unqualified snake_case name of this class.
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyident

    # Unqualified CamelCase name of this class.
    name_camelcase = get_name(cls)

    # Can it be? But it can.
    return pyident.convert_camelcase_to_snakecase(name_camelcase)

# ....................{ GETTERS ~ method                   }....................
@type_check
def get_method(cls: ClassType, method_name: str) -> CallableTypes:
    '''
    Possibly unbound method with the passed name defined by the passed class if
    any *or* raise an exception otherwise.

    If this method is:

    * A classmethod, this method is bound to this class.
    * A normal method, this method is unbound.
    * A static method, this method is unbound.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this method for.
    method_name : str
        Name of the method to be obtained.

    Returns
    ----------
    CallableTypes
        Possibly unbound method with this name.

    Raises
    ----------
    BetseMethodException
        If no such method is defined by this class.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # It's so simple, my eyes bleed.
    return objects.get_method(obj=cls, method_name=method_name)

# ....................{ ITERATORS                          }....................
@type_check
def iter_methods(cls: ClassType) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each method defined by
    the passed class (in ascending lexicographic order of method name).

    This generator *only* yields methods statically registered in the internal
    dictionary for this class (e.g., ``__dict__`` in unslotted classes),
    including:

    * Builtin methods, whose names are both prefixed and suffixed by ``__``.
    * Custom methods, whose names are *not* prefixed and suffixed by ``__``,
      including:
      * Custom standard methods.
      * Custom property methods (i.e., methods decorated by the builtin
        :func:`property` decorator).

    Parameters
    ----------
    cls : ClassType
        Class to iterate all methods of.

    Yields
    ----------
    (method_name, method_value)
        2-tuple of the name and value of each method bound to this object (in
        ascending lexicographic order of method name).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # Well, isn't that special?
    yield from objects.iter_methods(obj=cls)


@type_check
def iter_methods_matching(
    cls: ClassType, predicate: CallableTypes) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each method defined by
    the passed class whose method name matches the passed predicate (in
    ascending lexicographic order of method name).

    Parameters
    ----------
    cls: ClassType
        Class to yield all matching methods of.
    predicate : CallableTypes
        Callable iteratively passed the name of each method bound to this
        object, returning ``True`` only if that name matches this predicate.

    Yields
    ----------
    (method_name, method_value)
        2-tuple of the name and value of each matching method bound to this
        object (in ascending lexicographic order of method name).

    See Also
    ----------
    :func:`iter_methods`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # Well, isn't that special?
    yield from objects.iter_methods_matching(obj=cls, predicate=predicate)

# ....................{ DEFINERS                           }....................
@type_check
def define_class(
    # Mandatory parameters.
    class_name: str,

    # Optional parameters.
    class_attr_name_to_value: MappingType = {},
    base_classes: SequenceTypes = (),
) -> ClassType:
    '''
    Dynamically define a new class with the passed name subclassing all passed
    base classes and providing all passed class attributes (e.g., class
    variables, methods).

    Parameters
    ----------
    class_name : str
        Name of the class to be created.
    class_attr_name_to_value : MappingType
        Mapping from the name to the initial value of each class attribute
        (e.g., class variable, method) to declare this class to contain.
        Defaults to the empty dictionary, equivalent to declaring a class with
        the trivial body ``pass``.
    base_classes : optional[SequenceTypes]
        Sequence of all base classes to subclass this class from. Defaults to
        the empty tuple, equivalent to the 1-tuple ``(object,)`` containing only
        the root base class of all classes.

    Returns
    ----------
    ClassType
        Class dynamically defined with this name from these base classes and
        class attributes.
    '''

    # Thank you, bizarre 3-parameter variant of the type.__init__() method.
    return type(class_name, base_classes, class_attr_name_to_value)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.


from betse.science.cells import Cells
from betse.science.parameters import Parameters
from betse.science.sim import Simulator
from betse.science.tissue.handler import TissueHandler
from betse.util.io.log import logs
from betse.util.path import files, paths
from betse import ignition, pathtree

# Initialize the current application before doing anything else.
ignition.init()
# at the moment only the default config file can be used
config_filename = pathtree.CONFIG_DEFAULT_FILENAME

# Validate and localize such filename.
files.die_unless_file(config_filename)
_config_filename = config_filename
_config_basename = paths.get_basename(_config_filename)

logs.log_info(
    'Seeding simulation with configuration file "{}".'.format(
        _config_basename))

p = Parameters(config_filename=_config_filename)  # create an instance of Parameters
p.I_overlay = False  # force the current overlay to be null
sim = Simulator(p)  # create an instance of Simulator as it's needed by plotting objects

if p.sim_ECM is False:

    cells = Cells(p, worldtype='basic')  # create an instance of world
    cells.containsECM = False
    logs.log_info('Cell cluster is being created...')
    cells.makeWorld(p)  # call function to create the world

    # define the tissue and boundary profiles for plotting:
    logs.log_info('Defining tissue and boundary profiles...')
    sim.baseInit_all(cells, p)
    dyna = TissueHandler(sim, cells, p)
    dyna.tissueProfiles(sim, cells, p)

    cells.redo_gj(dyna, p)  # redo gap junctions to isolate different tissue types

    if p.fluid_flow is True or p.deformation is True:  # if user desires fluid flow:

        # make a laplacian and solver for discrete transfers on closed, irregular cell network
        logs.log_info('Creating cell network Poisson solver...')
        cells.graphLaplacian(p)

    cells.save_cluster(p)

    logs.log_info('Cell cluster creation complete!')


else:

    cells = Cells(p, worldtype='full')  # create an instance of world
    cells.containsECM = True
    logs.log_info('Cell cluster is being created...')
    cells.makeWorld(p)  # call function to create the world

    # define the tissue and boundary profiles for plotting:
    logs.log_info('Defining tissue and boundary profiles...')
    sim.baseInit_all(cells, p)
    dyna = TissueHandler(sim, cells, p)
    dyna.tissueProfiles(sim, cells, p)

    cells.redo_gj(dyna, p)  # redo gap junctions to isolate different tissue types

    # make a laplacian and solver for discrete transfers on closed, irregular cell network
    if p.fluid_flow is True or p.deformation is True:
        # loggers.log_info('Creating cell network Poisson solvers...')
        cells.graphLaplacian(p)

    cells.save_cluster(p)

    logs.log_info('Cell cluster creation complete!')


sim.sim_info_report(cells, p)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Utility functions simplifying access to :mod:`pytest` built-in ``request``
fixture.

The official documentation for this fixture is blatantly inadequate. Even were
this documentation adequate, however, this fixture's API is overly obtuse to an
absurd (and almost obfuscatory) degree. This module provides an intelligible
alternative to that API.

See Also
----------
https://pytest.org/latest/builtin.html#_pytest.python.FixtureRequest
    Official documentation for this fixture – such as it is.
'''

# Note that all functions defined below accepting a "request" fixture object
# assume the passed object to be of the expected type rather than validating
# this to be the case. Why?  Because this type resides in a private module that
# does *NOT* appear to be externally importable: e.g.,
#
#     >>> import _pytest.python.FixtureRequest as req
#     ImportError: cannot import name 'transfer_markers'

# ....................{ IMPORTS                            }....................
from pytest import Function
from betse.util.type import sequences
from betse.util.type.types import type_check, TestableTypes
from betse_test.exceptions import BetseTestFixtureException
from betse_test.util import pytests

# ....................{ CONSTANTS                          }....................
_DEFAULT_VALUE_NONE = object()
'''
Default value for the optional ``default_value`` argument accepted by the
:func:`get_fixture_param` function.

This value permits that function to distinguish between default values of value
``None`` and the lack of any default value altogether.
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_tested(request: '_pytest.python.FixtureRequest') -> None:
    '''
    Raise an exception unless the passed ``request`` fixture object was
    transitively requested by a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ----------
    BetseTestFixtureException
        If no test transitively requested this ``request`` fixture.

    See Also
    ----------
    :func:`is_tested`
        Further details.
    '''

    if not is_tested(request):
        raise BetseTestFixtureException(
            '"request" fixture transitively requested by no test '
            '(e.g., due to being requested by a file- or '
            'session-scoped fixture): {}'.format(request))

# ....................{ EXCEPTIONS ~ fixture               }....................
def die_unless_fixture(request: '_pytest.python.FixtureRequest') -> None:
    '''
    Raise an exception unless the passed ``request`` fixture object was directly
    requested by a fixture rather than a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ----------
    BetseTestFixtureException
        If a test rather than fixture requested this ``request`` fixture.
    '''

    if not is_fixture(request):
        raise BetseTestFixtureException(
            '"request" fixture requested by '
            'test rather than fixture: {}'.format(request))


def die_unless_fixture_parametrized(
    request: '_pytest.python.FixtureRequest') -> bool:
    '''
    Raise an exception unless the fixture requesting the passed ``request``
    fixture object was parametrized (either directly or indirectly).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ----------
    BetseTestFixtureException
        If either:
        * A test rather than fixture requested this ``request`` fixture.
        * A fixture that was _not_ parametrized requested this ``request``
          fixture.
    '''

    if not is_fixture_parametrized(request):
        raise BetseTestFixtureException(
            '"request" fixture requested by '
            'unparametrized fixture "{}".'.format(get_fixture_name(request)))

# ....................{ TESTERS                            }....................
def is_tested(request: '_pytest.python.FixtureRequest') -> bool:
    '''
    `True` only if the passed ``request`` fixture object was transitively
    requested by a test (e.g., directly by a test or indirectly by a test-scope
    fixture).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    bool
        `True` only if a test transitively requested this ``request`` fixture.
    '''

    return isinstance(request.node, Function)

# ....................{ TESTERS ~ fixture                  }....................
def is_fixture(request: '_pytest.python.FixtureRequest') -> bool:
    '''
    `True` only if the passed ``request`` fixture object was directly requested by
    a fixture rather than a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    bool
        `True` only if a fixture requested this ``request`` fixture.
    '''

    return request.fixturename is not None


def is_fixture_parametrized(request: '_pytest.python.FixtureRequest') -> bool:
    '''
    `True` only if the fixture requesting the passed ``request`` fixture object
    was parametrized (either directly or indirectly).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    bool
        `True` only if this fixture was parametrized.

    Raises
    ----------
    BetseTestFixtureException
        If a test rather than fixture requested this ``request`` fixture.
    '''

    # If this "request" fixture was requested by a test, raise an exception.
    die_unless_fixture(request)

    # Else, that fixture was itself requested by a fixture. Return whether on
    # not the latter was parametrized. py.test API: you fail again.
    return hasattr(request, 'param')

# ....................{ GETTERS                            }....................
def get_tested_name(request: '_pytest.python.FixtureRequest') -> str:
    '''
    Unqualified name of the current test (e.g., `test_cli_info`) if the passed
    ``request`` fixture object was either directly requested by that test *or*
    indirectly requested by a test-scope fixture.

    This function returns either:

    * If the current test is a test function, this function's unqualified name.
    * If the current test is a test method, this method's unqualified name.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    str
        Unqualified name of this test.

    Raises
    ----------
    BetseTestFixtureException
        If the passed ``request`` fixture object was *not* transitively requested
        by a test (e.g., was requested by a file- or session-scope fixture), in
        which case no test exists to obtain the name of.
    '''

    # Raise an exception unless a test transitively requested this fixture.
    die_unless_tested(request)

    # No one would ever think of trying this. Except one man did.
    return request.node.name

# ....................{ GETTERS ~ fixture                  }....................
def get_fixture_name(request: '_pytest.python.FixtureRequest') -> str:
    '''
    Name of the current fixture that explicitly requested the passed ``request``
    fixture object.

    If this ``request`` fixture was requested by a test rather than a fixture,
    an exception is raised.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    str
        Name of such fixture.
    '''

    # If this "request" fixture was requested by a test, raise an exception.
    die_unless_fixture(request)

    # Else, that fixture was itself requested by a fixture. Return the name of
    # the latter.
    return request.fixturename


@type_check
def get_fixture_param(
    request,
    default_value: object = _DEFAULT_VALUE_NONE,
    type_expected: TestableTypes = None,
) -> object:
    '''
    Parameter parametrizing the fixture requesting the passed ``request`` fixture,
    optionally defaulted to the passed default value if the former fixture is
    unparametrized _and_ optinionally validated to be an instance of the passed
    type or tuple of types.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    default_value: optional[object]
        Default value to return if this fixture is unparametrized. Defaults
        to :data:`_DEFAULT_VALUE_NONE`, in which case an exception is raised
        instead if this fixture is unparametrized.
    type_expected : optional[TestableTypes]
        Class or tuple of classes to validate this object to be an instance of.
        If this object is _not_ an instance of this class or classes, an
        exception is raised. Defaults to `None`, in which case this object is
        returned unvalidated.

    Returns
    ----------
    object
        Parameter parametrizing this fixture.

    Raises
    ----------
    BetseTestFixtureException
        If either:
        * A test rather than fixture requested this ``request`` fixture.
        * An unparametrized fixture requested this ``request`` fixture.
        * A fixture parametrized by an instance of a class _not_ the passed
          `check_type` requested this ``request`` fixture.
    '''

    # Parameter parametrizing the fixture requesting this "request" fixture
    fixture_param = None

    # If the caller passed a default value *AND* this fixture is unparametrized,
    # default this parameter to this value.
    if (default_value is not _DEFAULT_VALUE_NONE and
        not is_fixture_parametrized(request)):
        fixture_param = default_value
    # Else...
    else:
        # Raise an exception unless this fixture is parametrized.
        die_unless_fixture_parametrized(request)

        # Parameter parametrizing this fixture.
        fixture_param = request.param

    # Else, that fixture was parametrized. If the caller requested that
    # parametrization be validated, do so.
    if type_expected is not None and not isinstance(
        fixture_param, type_expected):
        raise BetseTestFixtureException(
            'Fixture "{}" parametrization not of type "{}": {!r}'.format(
                get_fixture_name(request), type_expected, fixture_param))

    # Return this parameter.
    return fixture_param

# ....................{ GETTERS ~ request                  }....................
#FIXME: This function internally calls the now-deprecated
#request.getfuncargvalue() function, which has since been replaced by the
#request.getfixturevalue() function. Sadly, the latter is unavailable in the
#most recent stable version of "pytest" available under my Linux distribution
#(Gentoo for the win) and hence ignored for the moment.
#
#Revisit this in early 2017, please.

@type_check
def get_requested_fixture(
    request,
    fixture_name: str,
    type_expected: TestableTypes = None,
) -> object:
    '''
    Object returned by the fixture with the passed name transitively requested
    by the current test requesting the passed ``request`` fixture object,
    optinionally validated to be an instance of the passed type or tuple of
    types.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name: str
        Name of the fixture to return the object returned by that fixture.
    type_expected : optional[TestableTypes]
        Class or tuple of classes to validate this object to be an instance of.
        If this object is _not_ an instance of this class or classes, an
        exception is raised. Defaults to `None`, in which case this object is
        returned unvalidated.

    Returns
    ----------
    _pytest.python.FixtureDef
        Object returned by this fixture.

    Raises
    ----------
    `pytests.get_fixture_lookup_error_type()`
        If this fixture is either unavailable _or_ is available but
        unretrievable (e.g., this fixture is the child of the current fixture).
    '''

    # Object returned by this fixture. Terrible Function Names Part XIXI: the
    # endless journey endures.
    fixture_object = request.getfuncargvalue(fixture_name)

    # If the caller requested this object be validated, do so.
    if type_expected is not None and not isinstance(
        fixture_object, type_expected):
        raise BetseTestFixtureException(
            'Fixture "{}" object not of type "{}": {!r}'.format(
                fixture_name, type_expected, fixture_object))

    # Return this object.
    return fixture_object


def get_requested_fixture_or_none(
    request: '_pytest.python.FixtureRequest',
    fixture_name: str,
) -> '_pytest.python.FixtureDef':
    '''
    Fixture with the passed name transitively requested by the current test if
    this fixture is both available and retrievable _or_ `None` otherwise,
    inspected from the passed ``request`` fixture object.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name: str
        Name of the fixture to return.

    Returns
    ----------
    _pytest.python.FixtureDef or None
        This fixture if this fixture is both available and retrievable _or_
        `None` otherwise.
    '''

    # Class of all "FixtureLookupError" exceptions raised by "pytest".
    FixtureLookupError = pytests.get_fixture_lookup_error_type()

    # Attempt to retrieve and return this fixture for this request.
    try:
        return get_requested_fixture(request, fixture_name)
    # If no such fixture exists, return None.
    except FixtureLookupError:
        return None

# ....................{ GETTERS ~ request : name           }....................
def get_requested_fixture_names(
    request: '_pytest.python.FixtureRequest') -> list:
    '''
    List of the names of all fixtures transitively requested by the current test
    excluding the fixture that explicitly requested the passed ``request`` fixture
    object if any, inspected from that ``request`` fixture object.

    This list includes the names of all fixtures that are:

    * Directly requested by this test.
    * Directly or indirectly requested by fixtures requested by this test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    list
        List of all such names.
    '''

    # List of the names of all fixtures transitively required by this test,
    # including the fixture that explicitly requested this "request" fixture if
    # any. Since the original list is guaranteed *NOT* to be modified below,
    # this list need not be copied from the original list to preserve the
    # latter.
    #
    # This attribute is undocumented. Why is this attribute undocumented? Work
    # with me here, people.
    fixture_names = request.fixturenames

    # If a fixture explicitly requested this "request" fixture...
    if is_fixture(request):
        # Name of this fixture.
        omit_fixture_name = get_fixture_name(request)

        # Exclude this name from this list *WITHOUT* modifying the original
        # list, as doing so would subsequently raise "KeyError" exceptions.
        fixture_names = sequences.omit_item(fixture_names, omit_fixture_name)
        assert omit_fixture_name not in fixture_names

    # Return these fixture names.
    return fixture_names

# ....................{ GETTERS ~ request : name : prefix  }....................
def get_requested_fixture_name_prefixed_by(
    request: '_pytest.python.FixtureRequest',
    fixture_name_prefix: str,
) -> str:
    '''
    Name prefixed by the passed prefix of the single fixture transitively
    requested by the current test, inspected from the passed ``request`` fixture
    object.

    If either no such fixture _or_ more than one such fixture exist, an
    exception is raised.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name_prefix: str
        String prefixing the fixture name to be returned.

    Returns
    ----------
    str
        Such name.

    See Also
    ----------
    :func:`get_requested_fixture_names`
        Further details.
    '''

    # List of the names of all fixtures prefixed by this prefix.
    prefixed_fixture_names = get_requested_fixture_names_prefixed_by(
        request, fixture_name_prefix)

    # Number of such fixtures.
    prefixed_fixture_count = len(prefixed_fixture_names)

    # If either no or more than one such fixtures exist, raise an exception.
    if prefixed_fixture_count != 1:
        # Exception message to be raised.
        exception_message = None

        if prefixed_fixture_count == 0:
            exception_message = (
                'No fixture prefixed by "{}" requested by this test.'.format(
                    fixture_name_prefix))
        else:
            exception_message = (
                'Multiple fixtures prefixed by "{}" '
                'requested by this test: {}'.format(
                    fixture_name_prefix, prefixed_fixture_names))

        # Raise this exception with this message.
        raise BetseTestFixtureException(exception_message)

    # Else, return the single such fixture name.
    return prefixed_fixture_names[0]


def get_requested_fixture_names_prefixed_by(
    request: '_pytest.python.FixtureRequest',
    fixture_name_prefix: str,
) -> list:
    '''
    List of all names prefixed by the passed prefix of all fixtures transitively
    requested by the current test, inspected from the passed ``request`` fixture
    object.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name_prefix: str
        String prefixing all fixture names in the returned list.

    Returns
    ----------
    list
        List of all such names.

    See Also
    ----------
    get_requested_fixture_names
        Further details.
    '''

    return sequences.get_items_prefixed_by(
        sequence=get_requested_fixture_names(request),
        item_prefix=fixture_name_prefix,
    )



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all YAML-backed simulation configuration subclasses
encapsulating on-disk lists and list items.
'''

# ....................{ IMPORTS                            }....................
from abc import abstractmethod
from betse.lib.yaml.yamlalias import yaml_alias
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.util.type.cls import classes
from betse.util.type.obj import objects
from betse.util.type.types import (
    type_check, ClassType, SequenceTypes)
from collections import MutableSequence

# ....................{ SUPERCLASSES ~ list item           }....................
class YamlListItemABC(YamlABC):
    '''
    Abstract base class of all YAML-backed list item subclasses, each instance
    of which is intended to be added to a parent :class:`YamlList` container.

    Each such instance may technically encapsulate any valid YAML type (e.g.,
    :class:`int`, :class:`str`) but typically encapsulates a YAML dictionary of
    related key-value pairs (e.g., animation settings, tissue profile).
    '''

    # ..................{ MAKERS                             }..................
    @classmethod
    def make_list(cls, *args, **kwargs) -> (
        'betse.lib.yaml.abc.yamllistabc.YamlList'):
        '''
        Create and return a low-level YAML-backed list containing only items of
        this specific subclass type, both loaded from and savable back into a
        YAML-formatted file.

        Parameters
        ----------
        All parameters are passed as is to the :meth:`YamlList.__init__` method.
        '''

        return YamlList(*args, item_type=cls, **kwargs)

    # ..................{ MAKERS ~ abstract                  }..................
    # Subclasses are required to implement the following abstract class methods.

    @classmethod
    @abstractmethod
    def make_default(
        cls,
        yaml_list: 'betse.lib.yaml.abc.yamllistabc.YamlList',
    ) -> 'betse.lib.yaml.abc.yamllistabc.YamlListItemABC':
        '''
        Create and return an instance of this subclass encapsulating a new
        dictionary containing default configuration settings.

        This class method is principally intended to be called by the
        :meth:`YamlList.append_default` method, appending this instance to an
        existing parent list of similar instances.

        Parameters
        ----------
        yaml_list : YamlList
            Parent list instantiating this instance.
        '''

        pass


class YamlListItemTypedABC(YamlListItemABC):
    '''
    Abstract base class of all low-level YAML-backed typed list item subclasses,
    each instance of which encapsulates a YAML dictionary whose keys define the
    type and name of this item intended to be added to a :class:`YamlList`
    container.

    Attributes
    ----------
    is_enabled : bool
        ``True`` only if this list item is enabled.
    name : str
        Lowercase alphanumeric string uniquely identifying the type of this
        list item (e.g., ``voltage_membrane``, signifying a transmembrane
        voltage list item). See each ``type`` key of the corresponding list in
        the default simulation configuration file for further commentary.
    '''

    # ..................{ ALIASES                            }..................
    is_enabled = yaml_alias("['enabled']", bool)
    name       = yaml_alias("['type']", str)

# ....................{ SUPERCLASSES ~ list                }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: To avoid diamond inheritance issues, neither this subclass nor any
# application-specific superclasses of this subclass (e.g., "YamlABC") should
# override any methods already defined by the "MutableSequence" superclass. This
# includes but is *NOT* limited to the public "MutableSequence" methods:
# append(), clear(), count(), extend(), index(), insert(), mro(), pop(),
# register(), remove(), and reverse().
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
class YamlList(YamlABC, MutableSequence):
    '''
    YAML-backed list subconfiguration both loaded from and saved back into a
    YAML-formatted file.

    Each item of this list is a :class:`YamlListItemABC` instance encapsulating
    the corresponding YAML-backed list item. Each such item may technically
    encapsulate any valid YAML type (e.g., :class:`int`, :class:`str`) but
    typically encapsulates a YAML dictionary of related key-value pairs (e.g.,
    animation settings, tissue profile).

    Attributes
    ----------
    _confs_wrap : list
        High-level list of each instance of the :attr:`_item_type` subclass
        encapsulating each dictionary in the low-level :attr:`_conf` list.
    _item_type : ClassType
        Subclass of the :class:`YamlListItemABC` abstract base class with
        which to instantiate each wrapper in the high-level :attr:`_confs_wrap`
        list encapsulating each dictionary in the low-level :attr:`_conf` list.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, item_type: ClassType) -> None:
        '''
        Initialize this low-level YAML-backed list.

        Attributes
        ----------
        item_type : ClassType
            Subclass of the :class:`YamlListItemABC` abstract base class
            with which to encapsulate each low-level YAML-backed list item in
            the passed ``confs`` list. For each such item, a new instance of
            this subclass is appended to the internal :attr:`_confs_wrap` list
            storing these instances.
        '''

        # Initialize our superclass.
        super().__init__()

        # If the passed type is *NOT* a high-level YAML-backed list item
        # subclass, raise an exception.
        classes.die_unless_subclass(
            subclass=item_type, superclass=YamlListItemABC)

        # Classify all passed parameters.
        self._item_type = item_type

        # Nullify all remaining parameters for safety.
        self._confs_wrap = None

    # ..................{ SUPERCLASS ~ YamlABC               }..................
    # Concrete methods provided by our YamlABC superclass.

    @type_check
    def load(self, conf: SequenceTypes) -> (
        'betse.lib.yaml.abc.yamllistabc.YamlList'):

        # Load our superclass with the passed sequence.
        super().load(conf=conf)

        # Initialize a high-level list of high-level wrappers.
        self._confs_wrap = []

        # For each low-level dictionary in this low-level sequence...
        for conf_yaml in conf:
            # High-level wrapper of the caller-specified type.
            conf_wrap = self._item_type()
            conf_wrap.load(conf=conf_yaml)

            # Wrap this dictionary with this wrapper.
            self._confs_wrap.append(conf_wrap)

        # Return this configuration for convenience.
        return self


    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Nullify this high-level list of high-level wrappers.
        self._confs_wrap = None

    # ..................{ SUPERCLASS ~ MutableSequence       }..................
    # Abstract methods required by our MutableSequence superclass.

    def __len__(self):
        return len(self._confs_wrap)


    @type_check
    def __delitem__(self, index: int) -> None:
        del self._conf[index]
        del self._confs_wrap[index]


    @type_check
    def __getitem__(self, index: int) -> YamlListItemABC:
        return self._confs_wrap[index]


    @type_check
    def __setitem__(self, index: int, value: YamlListItemABC) -> None:
        '''
        Set the low-level YAML-backed list item with the passed 0-based index to
        the passed such item.
        '''

        # If the passed object is *NOT* an instance of the caller-defined type,
        # raise an exception. This type is specified at initialization time and
        # hence cannot be type-checked via a method annotation.
        objects.die_unless_instance(obj=value, cls=self._item_type)

        # Set the high-level list item with this index to this object.
        self._confs_wrap[index] = value

        # Set the low-level list item with this index to this object's
        # underlying YAML-backed dictionary.
        self._conf[index] = value.conf


    @type_check
    def insert(self, index: int, value: YamlListItemABC) -> None:
        '''
        Insert the passed low-level YAML-backed list item immediately *before*
        the item with the passed 0-based index.
        '''

        # Raise an exception unless the passed object is an instance of the
        # desired API, which is specified at initialization time and hence
        # cannot be type checked above by a method annotation.
        objects.die_unless_instance(obj=value, cls=self._item_type)

        # Insert this object *BEFORE* the high-level list item with this index.
        self._confs_wrap.insert(index, value)

        # Insert this object's underlying YAML-backed dictionary *BEFORE* the
        # low-level list item with this index.
        self._conf.insert(index, value.conf)

    # ..................{ APPENDERS                          }..................
    def append_default(self) -> YamlListItemABC:
        '''
        Append a new low-level YAML-backed list item initialized to default
        values suitable for this list and return this item.

        Returns
        ----------
        YamlListItemABC
            Instance of the :attr:`_item_type` subclass appended to the
            high-level :attr:`_confs_wrap` list, encapsulating the new list item
            appended to the low-level :attr:`_conf` list.
        '''

        # New simulation configuration list item specific to this list.
        conf_wrap = self._item_type.make_default(self)

        # Append this list item to this list.
        self.append(conf_wrap)

        # Return this list item.
        return conf_wrap

    # ....................{ GETTERS                            }....................
    @type_check
    def get_item_name_unique(self, name_format: str) -> str:
        '''
        Identifier formatted according to the passed format specifier,
        guaranteed to be unique across all items of this list.

        By duck typing, each item of this list is required to define a ``name``
        instance variable whose value is the YAML-backed name of that item; if
        this is *not* the case, an exception is raised.

        This method then returns an identifier suitable for use by the caller as
        the value of the ``name`` instance variable for a newly created item of
        this list. Hence, callers tend to call this method from subclass
        :meth:`make_default` implementations.

        Parameters
        ----------
        name_format : str
            Format specifier containing a ``{}`` substring (e.g.,
            ``tissue ({})``), iteratively interpolated by this function with an
            arbitrary integer to produce this unique list item name.

        Returns
        ----------
        str
            Item name unique to this list matching this format.
        '''

        # Arbitrary unique identifier with which to uniquify (i.e., guarantee
        # the uniqueness of) the name of a new item in this list, defaulting to the
        # number of existing items in this list.
        yaml_list_item_id = len(self._confs_wrap)

        # Name of this tissue profile, unique in this list.
        yaml_list_item_name = None

        # While this name is *NOT* unique in this list, iteratively (re)search
        # this list until obtaining a unique name. This iteration is guaranteed
        # to (eventually) terminate successfully with a unique name.
        while True:
            yaml_list_item_name = name_format.format(yaml_list_item_id)

            # For each existing item of this list...
            for yaml_list_item_other in self._confs_wrap:
                # If this item already has this name, this name is non-unique.
                # In this case, increment this identifier, format a new name via
                # this identifier, and repeat this search.
                if yaml_list_item_name == yaml_list_item_other.name:
                    yaml_list_item_id += 1
                    break
            # If no item already has this name, this name is unique. In that
            # case, cease searching.
            else:
                break

        # Return this name.
        return yaml_list_item_name



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level classes aggregating all parameters pertaining to simulation events.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseSimConfException
from betse.science.tissue.event.tiseveabc import SimEventPulseABC
from betse.science.math import toolbox
from betse.util.io.log import logs
from betse.util.type.types import type_check, NoneType, NumericSimpleTypes

# ....................{ SUBCLASSES                         }....................
class SimEventPulseVoltage(SimEventPulseABC):
    '''
    Event applying a directed voltage to the environmental boundary for some
    range of simulation time steps.

    A positive voltage will be applied to one boundary edge during this period;
    a negative voltage will be applied to another.

    Attributes
    ----------------------------
    peak_voltage : NumericSimpleTypes
        Maximum voltage (V) to be applied.
    positive_voltage_boundary : str
        Character identifying the boundary edge to apply this positive voltage
        to. Valid values include:
        * ``T``, the top boundary.
        * ``B``, the bottom boundary.
        * ``L``, the left boundary.
        * ``R``, the right boundary.
    negative_voltage_boundary : str
        Character identifying the boundary edge to apply this negative voltage
        to. Valid values are as for ``side_positive_voltage`` above.
    '''

    # ..................{ PUBLIC                             }..................
    @type_check
    def __init__(
        self,
        start_time: NumericSimpleTypes,
        stop_time: NumericSimpleTypes,
        step_rate: NumericSimpleTypes,
        peak_voltage: NumericSimpleTypes,
        positive_voltage_boundary: str,
        negative_voltage_boundary: str,
    ) -> None:

        # Initialize our superclass with some passed parameters.
        super().__init__(
            start_time=start_time,
            stop_time=stop_time,
            step_rate=step_rate,
        )

        # Classify all remaining parameters.
        self.peak_voltage = peak_voltage
        self.positive_voltage_boundary = positive_voltage_boundary
        self.negative_voltage_boundary = negative_voltage_boundary


    @type_check
    def fire(self, sim: 'betse.science.sim.Simulator', t: NumericSimpleTypes) -> None:

        effector = toolbox.pulse(
            t, self.start_time, self.stop_time, self.step_rate)

        sim.bound_V[self.positive_voltage_boundary] = (
             self.peak_voltage * effector)
        sim.bound_V[self.negative_voltage_boundary] = (
            -self.peak_voltage * effector)

# ....................{ MAKERS                             }....................
@type_check
def make(p: 'betse.science.parameters.Parameters') -> (SimEventPulseVoltage, NoneType):
    '''
    Create and return a new :class:`SimEventPulseVoltage` instance if enabled by the
    passed simulation configuration *or* ``None`` otherwise.

    Parameters
    ----------------------------
    p : Parameters
        Current simulation configuration.

    Returns
    ----------------------------
    SimEventPulseVoltage, NoneType
        Either:
        * If enabled by the passed simulation configuration, a new
          :class:`SimEventPulseVoltage` instance.
        * Else, ``None``.
    '''

    # Object to be returned, defaulting to nothing.
    event = None

    # If this event is enabled, create an instance of this class.
    aev = p._conf['apply external voltage']

    if bool(aev['event happens']):
        # If extracellular spaces are enabled, parse this event.
        if p.is_ecm:
            event = SimEventPulseVoltage(
                start_time=float(aev['change start']),
                stop_time=float(aev['change finish']),
                step_rate=float(aev['change rate']),
                peak_voltage=float(aev['peak voltage']),
                positive_voltage_boundary=(
                    _convert_boundary_str_to_char(
                        aev['positive voltage boundary'])),
                negative_voltage_boundary=(
                    _convert_boundary_str_to_char(
                        aev['negative voltage boundary'])),
            )
        # Else, log a non-fatal warning.
        else:
            logs.log_warning(
                'Ignoring voltage event, '
                'as extracellular spaces are disabled.')

    return event

# ....................{ CONVERTERS                         }....................
#FIXME: Utter rubbish. Refactor this function to leverage a private global
#dictionary constant instead: e.g.,
#
#    BOUNDARY_STR_TO_CHAR = {
#        'top':    'T',
#        'bottom': 'B',
#        'left':   'L',
#        'right':  'R',
#    }
#    boundary_char = BOUNDARY_STR_TO_CHAR.get(side, None)
#    if boundary_char is None:
#        raise BetseSimConfException(
#            'Boundary edge "{}" unrecognized.'.format(side))
#    return boundary_char
#
#The sane means of doing so would be to refactor the implementations of the
#"positive_voltage_boundary" and "positive_voltage_boundary" attributes above
#into @property_cached-style properties. Possibly? Or not. That does rather
#seem like overkill. The current function-based approach is probably superior.
@type_check
def _convert_boundary_str_to_char(side: str) -> str:
    '''
    Convert the passed human-readable string constant identifying an
    environmental boundary edge (e.g., `top`) into the corresponding
    machine-readable character constant (e.g., `T`).
    '''

    if   side == 'top':    return 'T'
    elif side == 'bottom': return 'B'
    elif side == 'left':   return 'L'
    elif side == 'right':  return 'R'
    else:
        raise BetseSimConfException(
            'Boundary edge "{}" unrecognized.'.format(side))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level logging configuration.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific modules
# must *ALWAYS* be able to safely import from this module at any level, these
# circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import logging, os, sys
from betse import metadata, pathtree
from betse.util.io.log.logenum import LogLevel
from betse.util.io.log.logfilter import (
    LogFilterThirdPartyDebug, LogFilterMoreThanInfo)
from betse.util.io.log.logformat import LogFormatterWrap
from betse.util.io.log.loghandle import LogHandlerFileRotateSafe
from betse.util.type.types import type_check
from logging import Handler, RootLogger, StreamHandler
from os import path

# ....................{ GLOBALS                            }....................
# See below for utility functions accessing this singleton.
_config = None
'''
Singleton logging configuration for the current Python process.

This configuration provides access to root logger handlers. In particular, this
simplifies modification of logging levels at runtime (e.g., in response to
command-line arguments or configuration file settings).
'''

# ....................{ CONFIG                             }....................
#FIXME: Update docstring to reflect the new default configuration.
class LogConfig(object):
    '''
    BETSE-specific logging configuration.

    This configuration defines sensible default handlers for the root logger,
    which callers may customize (e.g., according to user-defined settings) by
    calling the appropriate getters.

    Caveats
    ----------
    Since this class' :meth:`__init__` method may raise exceptions, this class
    should be instantiated at application startup by an explicit call to the
    module-level :func:`init` function _after_ establishing default exception
    handling. Hence, this class is *not* instantiated at the end of this module.

    Default Settings
    ----------
    All loggers will implicitly propagate messages to the root logger configured
    by this class, whose output will be:

    * Formatted in a timestamped manner detailing the point of origin (e.g.,
      "[2016-04-03 22:02:47] betse ERROR (util.py:50): File not found.").
    * Labelled as the current logger's name, defaulting to `root`. Since this
      is _not_ a terribly descriptive name, callers are encouraged to replace
      this by an application-specific name.
    * Printed to standard error if the logging level for this output is either
      ``WARNING``, ``ERROR``, or ``CRITICAL``.
    * Printed to standard output if the logging level for this output is
      ``INFO``. Together with the prior item, this suggests that output with a
      logging level of ``DEBUG`` will *not* be printed by default.
    * Appended to the user-specific file specified by the
      :func:`pathtree.get_log_default_filename` function, whose:
        * Level defaults to :data:`logger.ALL`. Hence, *all* messages will be
          logged by default, including low-level debug messages. (This is
          helpful for debugging client-side errors.)
      * Contents will be automatically rotated on exceeding a sensible filesize
        (e.g., 16Kb).

    If the default log levels are undesirable, consider subsequently calling
    such logger's `set_level()` method. Since a desired log level is typically
    unavailable until *after* parsing CLI arguments and/or configuration file
    settings *AND* since a logger is required before such level becomes
    available, this function assumes a sane interim default.

    Attributes
    ----------
    _filename : str
        Absolute or relative path of the file logged to by the file handler,
        defaulting to :func:`pathtree.get_log_default_filename`.
    _logger_root : Logger
        Root logger.
    _logger_root_handler_file : Handler
        Root logger handler appending to the current logfile.
    _logger_root_handler_stderr : Handler
        Root logger handler printing to standard error.
    _logger_root_handler_stdout : Handler
        Root logger handler printing to standard output.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self):
        '''
        Initialize this logging configuration as documented by the class
        docstring.
        '''

        # Avoid circular import dependencies.

        # Initialize the superclass.
        super().__init__()

        # Initialize all non-property attributes to sane defaults. To avoid
        # chicken-and-egg issues, properties should *NOT* be set here.
        self._filename = pathtree.get_log_default_filename()
        self._logger_root = None
        self._logger_root_handler_file = None
        self._logger_root_handler_stderr = None
        self._logger_root_handler_stdout = None

        # Initialize the root logger.
        self._init_logger_root()

        # Initialize root logger handlers *AFTER* the root logger, as the former
        # explicitly add themselves to the latter.
        self._init_logger_root_handler_std()
        self._init_logger_root_handler_file()

        # Redirect all warnings through the logging framewark *AFTER*
        # successfully performing the above initialization.
        logging.captureWarnings(True)


    def _init_logger_root(self) -> None:
        '''
        Initialize the root logger.

        For safety, this function removes all previously initialized handlers
        from this logger.
        '''

        # Root logger.
        self._logger_root = logging.getLogger()

        # For uniqueness, change the name of the root logger to that of our
        # top-level package "betse" from its ambiguous default "root".
        self._logger_root.name = metadata.PACKAGE_NAME

        # Instruct this logger to entertain all log requests, ensuring these
        # requests will be delegated to the handlers defined below. By default,
        # this logger ignores all log requests with level less than "WARNING",
        # preventing handlers from receiving these requests.
        self._logger_root.setLevel(LogLevel.ALL)

        # For safety, remove all existing handlers from the root logger. While
        # this should *NEVER* be the case for conventional BETSE runs, this is
        # usually the case for functional BETSE tests *NOT* parallelized by
        # "xdist" and hence running in the same Python process. For safety,
        # iterate over a shallow copy of the list of handlers to be removed
        # rather than the actual list being modified here.
        for root_handler in list(self._logger_root.handlers):
            self._logger_root.removeHandler(root_handler)


    def _init_logger_root_handler_std(self) -> None:
        '''
        Initialize root logger handlers redirecting log messages to the standard
        stdout and stderr file handles.
        '''

        # Avoid circular import dependencies.
        from betse.util.path.command import cmds

        # Initialize the stdout handler to:
        #
        # * Log only informational messages by default.
        # * Unconditionally ignore all warning and error messages, which the
        #   stderr handler already logs.
        #
        # Sadly, the "StreamHandler" constructor does *NOT* accept the customary
        # "level" attribute accepted by its superclass constructor.
        self._logger_root_handler_stdout = StreamHandler(sys.stdout)
        self._logger_root_handler_stdout.setLevel(LogLevel.INFO)
        self._logger_root_handler_stdout.addFilter(LogFilterMoreThanInfo())

        # Initialize the stderr handler to:
        #
        # * Log only warning and error messages by default.
        # * Unconditionally ignore all informational and debug messages, which
        #   the stdout handler already logs.
        self._logger_root_handler_stderr = StreamHandler(sys.stderr)
        self._logger_root_handler_stderr.setLevel(LogLevel.WARNING)

        # Avoid printing third-party debug messages to the terminal.
        self._logger_root_handler_stdout.addFilter(LogFilterThirdPartyDebug())
        self._logger_root_handler_stderr.addFilter(LogFilterThirdPartyDebug())

        #FIXME: Consider colourizing this format string.

        # Format standard output and error in the conventional way. For a list
        # of all available log record attributes, see:
        #
        #     https://docs.python.org/3/library/logging.html#logrecord-attributes
        #
        # Note that "{{" and "}}" substrings in format() strings escape literal
        # "{" and "}" characters, respectively.
        stream_format = '[{}] {{message}}'.format(cmds.get_current_basename())

        # Formatter for this format.
        stream_formatter = LogFormatterWrap(fmt=stream_format, style='{')

        # Assign these formatters to these handlers.
        self._logger_root_handler_stdout.setFormatter(stream_formatter)
        self._logger_root_handler_stderr.setFormatter(stream_formatter)

        # Register these handlers with the root logger.
        self._logger_root.addHandler(self._logger_root_handler_stdout)
        self._logger_root.addHandler(self._logger_root_handler_stderr)


    def _init_logger_root_handler_file(self) -> None:
        '''
        Initialize the root logger handler appending log messages to the
        currently open logfile file handle.

        This method is designed to be called multiple times, permitting the
        filename associated with this handler to be modified at runtime.
        '''

        # Avoid circular import dependencies.
        from betse.util.path.command import cmds
        from betse.util.type.numeric import ints

        # Absolute or relative path of the directory containing this file.
        file_dirname = path.dirname(self._filename)

        # Minimum level of messages to be log to disk, defaulting to "INFO".
        file_level = LogLevel.INFO

        # If this handler has already been created...
        if self._logger_root_handler_file is not None:
            # Preserve the previously set minimum level of messages to log.
            file_level = self._logger_root_handler_file.level

            # If the root logger has also already been created, remove this
            # handler from this root logger.
            if self._logger_root is not None:
                self._logger_root.removeHandler(self._logger_root_handler_file)

        # If the path of the directory containing this file is non-empty, create
        # this directory if needed. Note this path is empty when this filename
        # is a pure basename (e.g., when the "--log-file=my.log" option is
        # passed).
        #
        # For safety, this directory is created with standard low-level Python
        # functionality rather than our custom higher-level
        # dirs.make_parent_unless_dir() function. The latter logs this creation.
        # Due to the root logger having not been fully configured yet, calling
        # that function here would induce subtle errors or exceptions.
        if file_dirname:
            os.makedirs(file_dirname, exist_ok=True)

        # Root logger file handler, preconfigured as documented above.
        self._logger_root_handler_file = LogHandlerFileRotateSafe(
            filename=self._filename,

            # Append rather than overwrite this file.
            mode='a',

            # Defer opening this file in a just-in-time manner (i.e., until the
            # first call to this handler's emit() method is called to write the
            # first log via this handler). Why? Because (in no particular
            # order):
            #
            # * If the end user requests that *NO* messages be logged to disk
            #   (e.g., by passing the "--log-level=none" option), this file
            #   should *NEVER* be opened and hence created. The simplest means
            #   of doing so is simply to indefinitely defer opening this file.
            # * Doing so slightly reduces the likelihood (but *NOT* eliminate
            #   the possibility) of race conditions between multiple BETSE
            #   processes attempting to concurrently rotate the same logfile.
            delay=True,

            # Encode this file's contents as UTF-8.
            encoding='utf-8',

            # Maximum filesize in bytes at which to rotate this file, equivalent
            # to 1 MB.
            maxBytes=ints.MiB,

            # Maximum number of rotated logfiles to maintain.
            backupCount=8,
        )

        # Initialize this handler's level to the previously established level.
        self._logger_root_handler_file.setLevel(file_level)

        # Prevent third-party debug messages from being logged to disk.
        self._logger_root_handler_file.addFilter(LogFilterThirdPartyDebug())

        # Linux-style logfile format.
        #
        # Note that the "processName" attribute appears to *ALWAYS* expand to
        # "MainProcess", which is not terribly descriptive. Hence, the name of
        # the current process is manually embedded in this format.
        file_format = (
            '[{{asctime}}] '
            '{} {{levelname}} '
            '({{module}}.py:{{funcName}}():{{lineno}}) '
            '<PID {{process}}>:\n'
            '    {{message}}'.format(cmds.get_current_basename()))

        # Format this file according to this format.
        file_formatter = LogFormatterWrap(fmt=file_format, style='{')
        self._logger_root_handler_file.setFormatter(file_formatter)

        # Register this handler with the root logger.
        self._logger_root.addHandler(self._logger_root_handler_file)

    # ..................{ PROPERTIES ~ logger                }..................
    # Read-only properties prohibiting write access to external callers.

    @property
    def logger_root(self) -> RootLogger:
        '''
        **Root logger** (i.e., transitive parent of all other loggers).
        '''

        return self._logger_root

    # ..................{ PROPERTIES ~ handler               }..................
    @property
    def handler_file(self) -> Handler:
        '''
        Root logger handler appending to the current logfile if file logging is
        enabled *or* ``None`` otherwise.
        '''

        return self._logger_root_handler_file


    @property
    def handler_stderr(self) -> Handler:
        '''
        Root logger handler printing to standard error.
        '''

        return self._logger_root_handler_stderr


    @property
    def handler_stdout(self) -> Handler:
        '''
        Root logger handler printing to standard output.
        '''

        return self._logger_root_handler_stdout

    # ..................{ PROPERTIES ~ level                 }..................
    @property
    def file_level(self) -> LogLevel:
        '''
        Minimum level of messages to log to the file handler.
        '''

        return self._logger_root_handler_file.level


    @file_level.setter
    @type_check
    def file_level(self, file_level: LogLevel) -> None:
        '''
        Set the minimum level of messages to log to the file handler.
        '''

        self._logger_root_handler_file.setLevel(file_level)

    # ..................{ PROPERTIES ~ level : verbose       }..................
    @property
    def is_verbose(self) -> bool:
        '''
        ``True`` only if *all* messages are to be unconditionally logged to the
        stdout handler (and hence printed to stdout).

        Equivalently, this method returns ``True`` only if the logging level for
        the stdout handler is :attr:`LogLevel.ALL`.

        Note that this logging level is publicly retrievable by accessing the
        :attr:`handler_stdout.level` property.
        '''

        return self._logger_root_handler_stdout.level == LogLevel.ALL


    @is_verbose.setter
    @type_check
    def is_verbose(self, is_verbose: bool) -> None:
        '''
        Set the verbosity of the stdout handler.

        This method sets this handler's logging level to:

        * If the passed boolean is ``True``, :attr:`LogLevel.ALL` .
        * If the passed boolean is ``False``, :attr:`LogLevel.INFO`.
        '''

        # Convert the passed boolean to a logging level for the stdout handler.
        self._logger_root_handler_stdout.setLevel(
            LogLevel.ALL if is_verbose else LogLevel.INFO)
        # print('handler verbosity: {} ({})'.format(self._logger_root_handler_stdout.level, ALL))

    # ..................{ PROPERTIES ~ path                  }..................
    @property
    def filename(self) -> str:
        '''
        Absolute or relative path of the file logged to by the file handler.
        '''

        return self._filename


    @filename.setter
    @type_check
    def filename(self, filename: str) -> None:
        '''
        Set the absolute or relative path of the file logged to by the file
        handler.

        Due to flaws in the upstream :mod:`logging` API, this method necessarily
        destroys and recreates the current file handler.
        '''

        # If the passed filename is the same as the current filename, avoid
        # unnecessarily destroying and recreating the file handler. This is
        # technically a negligible optimization, but every little bit helps.
        if self._filename == filename:
            return

        # Classify this filename *BEFORE* recreating the file handler, which
        # accesses this variable.
        self._filename = filename

        # Destroy and recreate the file handler.
        self._init_logger_root_handler_file()

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Enable the default logging configuration for the active Python process.
    '''

    # Instantiate this singleton global with the requisite defaults.
    # print('Reinitializing logging.')
    global _config
    _config = LogConfig()

# ....................{ GETTERS                            }....................
def get() -> LogConfig:
    '''
    Singleton logging configuration for the active Python process.
    '''

    return _config


def get_metadata():
    '''
    Ordered dictionary synopsizing the current logging configuration.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.mapping.mapcls import OrderedArgsDict

    # Return this dictionary.
    return OrderedArgsDict(
        'file', _config.filename,
        'file level', _config.file_level.name.lower(),
        'verbose', str(_config.is_verbose).lower(),
    )



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all YAML-backed simulation configuration subclasses as
well as functionality pertaining to such classes.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.exceptions import BetseYamlException
from betse.lib.yaml import yamls
from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from betse.util.path.dirs import DirOverwritePolicy
from betse.util.type.iterators import empty_iterator
from betse.util.type.types import (
    type_check,
    IterableTypes,
    MappingOrSequenceTypes,
    NoneType,
)

# ....................{ SUPERCLASSES                       }....................
class YamlABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **YAML wrapper** (i.e., high-level object
    wrapping a low-level mapping or sequence of YAML-backed configuration
    settings) subclasses.

    Attributes
    ----------
    _conf : MappingOrSequenceOrNoneTypes
        Low-level container of related configuration settings loaded from and
        saved back to a YAML-formatted configuration file if the :meth:`load`
        method has been called *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this YAML wrapper.
        '''

        self._conf = None

    # ..................{ PROPERTIES ~ read-only             }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def is_loaded(self) -> bool:
        '''
        ``True`` only if this configuration is in the **loaded state** (i.e.,
        the :meth:`load` method has been called more recently than the
        :meth:`unload` method has been called).

        If ``True``, the :meth:`conf` property is safely accessible by callers.
        '''

        return self._conf is not None


    @property
    def conf(self) -> MappingOrSequenceTypes:
        '''
        Low-level container of related configuration settings loaded from and
        saved back to a YAML-formatted configuration file if the :meth:`load`
        method has been called more recently than the :meth:`unload` method has
        been called *or* raise an exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Else, this property is a container. Return this container.
        return self._conf

    # ..................{ EXCEPTIONS                         }..................
    def die_unless_loaded(self) -> None:
        '''
        Raise an exception unless this configuration is in the loaded state.

        Equivalently, this method raises an exception if this configuration is
        in the unloaded state.

        Raises
        ----------
        BetseYamlException
            If this configuration is in the unloaded state.

        See Also
        ----------
        :meth:`is_loaded`
            Further details.
        '''

        if not self.is_loaded:
            raise BetseYamlException(
                'YAML configuration not loaded '
                '(i.e., load() method not called yet).')

    # ..................{ LOADERS                            }..................
    @type_check
    def load(self, conf: MappingOrSequenceTypes) -> (
        'betse.lib.yaml.abc.yamlabc.YamlABC'):
        '''
        Associate this configuration with the passed YAML-backed container.

        Parameters
        ----------
        conf : MappingOrSequenceTypes
            Low-level container of related configuration settings loaded from
            and saved back to a YAML-formatted configuration file.

        Returns
        ----------
        YamlABC
            This YAML wrapper, returned for caller convenience.
        '''

        # Classify this container.
        self._conf = conf

        # Return this configuration for convenience.
        return self


    def unload(self) -> None:
        '''
        Deassociate this configuration from its YAML-backed container if any
        *or* reduce to a noop otherwise.

        This method serves as a low-level safety mechanism ensuring that the
        next access of the :meth:`conf` property will raise a human-readable
        exception.
        '''

        self._conf = None

# ....................{ SUPERCLASSES ~ file                }....................
class YamlFileABC(YamlABC):
    '''
    Abstract base class of all **YAML file wrapper** (i.e., high-level object
    wrapping a low-level mapping or sequence of YAML-backed configuration
    settings both loadable from and savable back to a YAML-formatted file)
    subclasses.

    Caveats
    ----------
    The ``conf_filename`` parameter accepted by most methods of this class
    (e.g., :meth:`load`, :meth:`write`) should be suffixed by a valid YAML
    filetype -- namely, either ``.yml`` or ``.yaml``. This class does *not*
    erforce this recommendation but does log non-fatal warnings where violated.

    External callers should *never* access the low-level :meth:`conf` property,
    as doing so violates type safety and value constraints. The configuration
    settings persisted by this property are safely retrievable and modifiable
    *only* via :func:`yaml_alias` data descriptors defined by subclasses.

    Attributes
    ----------
    _conf_basename : StrOrNoneTypes
        Basename of :attr:`_conf_filename` if non-``None`` *or* ``None``
        otherwise.
    _conf_dirname : StrOrNoneTypes
        Absolute dirname of :attr:`_conf_filename` if non-``None`` *or* ``None``
        otherwise.
    _conf_filename : StrOrNoneTypes
        Absolute filename of the low-level YAML-formatted configuration file
        from which this object was most recently deserialized if any *or*
        ``None`` otherwise.
    '''

    # ..................{ MAKERS                             }..................
    @classmethod
    @type_check
    def make(cls, conf_filename: str, *args, **kwargs) -> (
        'betse.lib.yaml.abc.yamlabc.YamlFileABC'):
        '''
        Create and return a YAML file wrapper of this subclass type,
        deserialized from the passed YAML-formatted file into a low-level
        mapping or sequence internally persisted in this wrapper.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative filename of the source file to be deserialized.

        All other parameters are passed as is to the :meth:`__init__` method.

        Returns
        ----------
        YamlFileABC
            This configuration, returned for convenience chaining in callers.
        '''

        return cls(*args, **kwargs).load(conf_filename)

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this YAML file wrapper in the **unload state** (i.e.,
        associated with *no* low-level YAML-formatted configuration file).
        '''

        # Initialize our superclass.
        super().__init__()

        # Initialize this simulation configuration in the unload state. To avoid
        # extraneous logging, the unload() method body is duplicated here.
        self._conf_basename = None
        self._conf_dirname = None
        self._conf_filename = None

    # ..................{ PROPERTIES ~ read-only             }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def conf_basename(self) -> str:
        '''
        Basename of :attr:`conf_filename` if a prior call to the :meth:`load`
        method successfully loaded such a file *or* raise an exception
        otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_basename


    @property
    def conf_filename(self) -> str:
        '''
        Absolute filename of the low-level YAML-formatted file from which this
        file was deserialized if a prior call to the :meth:`load` method
        successfully loaded such a file *or* raise an exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_filename


    @property
    def conf_dirname(self) -> str:
        '''
        Absolute direname of :attr:`conf_filename` if a prior call to the
        :meth:`load` method successfully loaded such a file *or* raise an
        exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_dirname

    # ..................{ LOADERS                            }..................
    #FIXME: Given the existence of the make() class method, we no longer need
    #either this or the superclass or any subclass methods to return themselves
    #anymore. That's a good thing, as the need to "return self" from each such
    #method has been the repeated source of issues throughout the codebase.
    @type_check
    def load(self, conf_filename: str) -> (
        'betse.lib.yaml.abc.yamlabc.YamlFileABC'):
        '''
        Deserialize the passed YAML-formatted file into a low-level mapping or
        sequence internally persisted in this wrapper.

        Usage
        ----------
        This method may be safely called multiple times with different files.
        When doing so, note that the current call to this method takes absolute
        precedence over all prior calls to this method -- which the former
        effectively nullifies. For safety, this method internally calls the
        :meth:`unload` method as needed (i.e., if some file has already been
        loaded). Ergo, that method need *not* be externally called to
        deassociate this wrapper from previously loaded files.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative filename of the source file to be deserialized.

        Returns
        ----------
        YamlFileABC
            This YAML file wrapper, returned for caller convenience.
        '''

        # If a file is already loaded, unload this file for safety. While doing
        # so effectively reduces to a noop with respect to this superclass,
        # subclass implementations of the unload() method may elect to perform
        # critical functionality that *MUST* be respected here.
        if self.is_loaded:
            self.unload()

        # Log this operation.
        logs.log_debug(
            'Loading YAML file "%s"...', pathnames.get_basename(conf_filename))

        # Low-level dictionary deserialized from this file.
        conf = yamls.load(filename=conf_filename)

        # Load this dictionary into our superclass.
        super().load(conf=conf)

        # Associate this object with this file *AFTER* successfully
        # deserializing this file.
        self._set_conf_filename(conf_filename)

        # Return this configuration for convenience.
        return self


    def unload(self) -> None:

        # If this file is actually loaded, log this operation.
        if self.is_loaded:
            logs.log_debug('Closing YAML file "%s"...', self._conf_basename)

        # Unload our superclass.
        super().unload()

        # Nullify all instance variables for safety.
        self._conf_basename = None
        self._conf_dirname = None
        self._conf_filename = None

    # ..................{ SAVERS                             }..................
    @type_check
    def save(self, conf_filename: str) -> None:
        '''
        Serialize the low-level dictionary internally stored in this object to
        the passed YAML-formatted simulation configuration file, replacing the
        prior contents of this file, *and* associate this object with this file.

        This method effectively implements the "Save As..." GUI metaphor.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative path of the target file to be serialized.
        '''

        # Log this operation.
        logs.log_debug('Saving YAML file "%s"...', self._conf_basename)

        # Save this dictionary to this file.
        yamls.save(
            container=self.conf,
            filename=conf_filename,
            is_overwritable=True,
        )

        # Absolute paths of the parent directories containing the current file
        # and the passed file, canonicalized to permit comparison below.
        src_dirname = pathnames.canonicalize(self.conf_dirname)
        trg_dirname = pathnames.canonicalize(
            pathnames.get_dirname(conf_filename))

        # If these paths differ, recursively copy all relative subdirectories
        # internally referenced and hence required by this file.
        if src_dirname != trg_dirname:
            # For the relative pathname of each such subdirectory...
            for conf_subdirname in self._iter_conf_subdirnames():
                # Absolute pathname of the old subdirectory.
                src_subdirname = pathnames.join(src_dirname, conf_subdirname)

                # Recursively copy from the old into the new subdirectory.
                dirs.copy_into_dir(
                    src_dirname=src_subdirname,
                    trg_dirname=trg_dirname,
                    overwrite_policy=DirOverwritePolicy.OVERWRITE,
                )

        # Associate this object with this file *AFTER* successfully copying to
        # this file and all external paths required by this file.
        self._set_conf_filename(conf_filename)


    @type_check
    def save_inplace(self) -> None:
        '''
        Serialize the low-level dictionary internally stored in this object to
        the current YAML-formatted simulation configuration file associated with
        this object, replacing the prior contents of this file.

        This method effectively implements the "Save" GUI metaphor.
        '''

        # Log this operation.
        logs.log_debug('Overwriting YAML file "%s"...', self._conf_basename)

        # Resave this dictionary to this file.
        yamls.save(
            container=self.conf,
            filename=self.conf_filename,
            is_overwritable=True,
        )

    # ..................{ SETTERS                            }..................
    @type_check
    def _set_conf_filename(self, conf_filename: str) -> None:
        '''
        Set the absolute path of the YAML-formatted file associated with this
        configuration.

        Design
        ----------
        To prevent external callers from unsafely setting this path, this setter
        is intentionally implemented as an manual setter rather than a more
        preferable :meth:`conf_filename` property setter.
        '''

        # Unique absolute filename of this file assigned *BEFORE* this file's
        # parent directory, ensuring the latter is non-empty.
        self._conf_filename = pathnames.canonicalize(conf_filename)

        # Basename of this file.
        self._conf_basename = pathnames.get_basename(self._conf_filename)

        # Unique absolute dirname of the parent directory of this file.
        self._conf_dirname = pathnames.get_dirname(self._conf_filename)

    # ..................{ SUBCLASS ~ optional                }..................
    # Methods intended to be optionally overriden by subclasses.

    def _iter_conf_subdirnames(self) -> IterableTypes:
        '''
        Generator yielding the pathname of each requisite direct subdirectory of
        the parent directory of the YAML-formatted file currently associated
        with this configuration, where "requisite" means internally referenced
        and hence required by this file.

        For safety, this generator excludes all direct subdirectories of this
        parent directory that are *not* internally referenced by this file.

        By default, this superclass method returns the empty generator.

        Yields
        ----------
        str
            Absolute or relative pathname of each such subdirectory.
        '''

        # If no file has been read, raise an exception.
        self.die_unless_loaded()

        # Default to the empty iterator.
        return empty_iterator()

# ....................{ TYPES                              }....................
# Intended for use in callable type validation.
YamlABCOrNoneTypes = (YamlABC, NoneType)
'''
Tuple of both the YAML-backed configuration type *and* the type of the singleton
``None`` object.
'''
